/*
 * ParametrizedTestSuiteTest.java
 * JUnit based test
 *
 * Created on 3. Mrz 2002, 18:47
 */

package de.tivano.junit;

import junit.framework.*;
import java.util.Enumeration;
import java.util.Properties;
import junit.framework.TestResult;
import junit.framework.AssertionFailedError;
import junit.framework.TestSuite;
import junit.framework.Test;
import java.lang.reflect.Method;
import junit.extensions.TestDecorator;
import java.io.InputStream;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Field;
import java.lang.reflect.Constructor;
import junit.framework.TestCase;
import java.util.Iterator;
import java.util.Set;
import java.util.HashSet;

/**
 *
 * @author Richard Kunze
 */
public class ParametrizedTestSuiteTest extends TestCase {
    
    public ParametrizedTestSuiteTest(java.lang.String testName) {
        super(testName);
    }
    
    public static void main(java.lang.String[] args) {
        junit.textui.TestRunner.run(suite());
    }
    
    public static Test suite() {
        TestSuite suite = new TestSuite(ParametrizedTestSuiteTest.class);
        
        return suite;
    }

    private Properties testProperties;
    
    
    /** Set up the fixture */
    public void setUp() throws IOException {
        InputStream data = getClass().getResourceAsStream(
           "/de/tivano/junit/ParametrizationTestDummy.properties");
        testProperties = new Properties();
        testProperties.load(data);
    }
    
    /** Test of warning method, of class de.tivano.junit.ParametrizedTestSuite. */
    public void testWarning() {
        ParametrizedTestSuite suite = new ParametrizedTestSuite(new Properties());
        
        Test warning = suite.warning("This is a warning");
        
        assertTrue("Expected the warning to be an instance of ParametrizedTestSuite.Warning", 
                   warning instanceof ParametrizedTestSuite.Warning);
        assertEquals("Test name:", "warning", ((TestCase)warning).getName());
        
        TestResult result = ((TestCase)warning).run();
        assertEquals("failure count:", 1, result.failureCount());
        assertEquals("error count:", 0, result.errorCount());
        TestFailure failure = (TestFailure)result.failures().nextElement();
        assertEquals("failure message:", "This is a warning", 
                     failure.thrownException().getMessage());
    }
    
    /** Test the {@link ParametrizedTestSuite#ParametrizedTestSuite(Properties)} constructor 
     *  with a non-empty properties object.
     */
    public void testParametrizedTestSuite_Properties() {
         ParametrizedTestSuite suite = new ParametrizedTestSuite(testProperties);
         assertEquals("Test suite name", null, suite.getName());
         assertEquals("Number of test cases", 0, suite.testCount());
   }
    
    /** Test the {@link ParametrizedTestSuite#ParametrizedTestSuite(Properties)} constructor 
     *  with an empty properties object.
     * The test expects that a warning is added to the test suite.
     */
    public void testParametrizedTestSuite_Properties_Empty() {
        ParametrizedTestSuite suite = new ParametrizedTestSuite(new Properties());
        assertEquals("Number of test cases", 1, suite.testCount());
        assertEquals("Expected the suite to contain only warnings",
                     ParametrizedTestSuite.Warning.class.getName(),
                     suite.testAt(0).getClass().getName());
        TestResult result = new TestResult();
        suite.testAt(0).run(result);
        Enumeration failures = result.failures();
        assertEquals("warning message:",
                     "Test data definition for test suite 'null' is empty!",
                     ((TestFailure)failures.nextElement()).thrownException().getMessage());        
    }
    
     /** Test the {@link ParametrizedTestSuite#ParametrizedTestSuite(Properties)} constructor 
     *  with an null properties object.
     * The test expects that a warning is added to the test suite.
     */
    public void testParametrizedTestSuite_Properties_Null() {
        ParametrizedTestSuite suite = new ParametrizedTestSuite((Properties)null);
        assertEquals("Number of test cases", 1, suite.testCount());
        assertEquals("Expected the suite to contain only warnings",
                     ParametrizedTestSuite.Warning.class.getName(),
                     suite.testAt(0).getClass().getName());
        TestResult result = new TestResult();
        suite.testAt(0).run(result);
        Enumeration failures = result.failures();
        assertEquals("warning message:",
                     "No test data specified for test suite 'null'!",
                     ((TestFailure)failures.nextElement()).thrownException().getMessage());        
    }
    
    /** Test the {@link ParametrizedTestSuite#ParametrizedTestSuite(String, Properties)} constructor 
     *  with a non-empty properties object.
     */
    public void testParametrizedTestSuite_String_Properties() {
         ParametrizedTestSuite suite = new ParametrizedTestSuite("some_name", testProperties);
         assertEquals("Test suite name", "some_name", suite.getName());
         assertEquals("Number of test cases", 0, suite.testCount());
   }
    
    /** Test the {@link ParametrizedTestSuite#ParametrizedTestSuite(Properties)} constructor 
     *  with an empty properties object.
     * The test expects that a warning is added to the test suite.
     */
    public void testParametrizedTestSuite_String_Properties_Empty() {
        ParametrizedTestSuite suite = new ParametrizedTestSuite("some_name", new Properties());
        assertEquals("Test suite name", "some_name", suite.getName());
        assertEquals("Number of test cases", 1, suite.testCount());
        assertEquals("Expected the suite to contain only warnings",
                     ParametrizedTestSuite.Warning.class.getName(),
                     suite.testAt(0).getClass().getName());
        TestResult result = new TestResult();
        Enumeration failures = result.failures();
        suite.testAt(0).run(result);
        assertEquals("warning message:",
                     "Test data definition for test suite 'some_name' is empty!",
                     ((TestFailure)failures.nextElement()).thrownException().getMessage());        
    }
    
     /** Test the {@link ParametrizedTestSuite#ParametrizedTestSuite(Properties)} constructor 
     *  with an null properties object.
     * The test expects that a warning is added to the test suite.
     */
    public void testParametrizedTestSuite_String_Properties_Null() {
        ParametrizedTestSuite suite = new ParametrizedTestSuite("some_name", (Properties)null);
        assertEquals("Test suite name", "some_name", suite.getName());
        assertEquals("Number of test cases", 1, suite.testCount());
        assertEquals("Expected the suite to contain only warnings",
                     ParametrizedTestSuite.Warning.class.getName(),
                     suite.testAt(0).getClass().getName());
        TestResult result = new TestResult();
        Enumeration failures = result.failures();
        suite.testAt(0).run(result);
        assertEquals("warning message:",
                     "No test data specified for test suite 'some_name'!",
                     ((TestFailure)failures.nextElement()).thrownException().getMessage());        
    }
    
   /** Test the {@link ParametrizedTestSuite#ParametrizedTestSuite(String)} constructor 
     *  with a name that has no matching properties data.
     * The test expects that a warning is added to the test suite.
     */
    public void testParametrizedTestSuite_String_NoData() {
        ParametrizedTestSuite suite = 
            new ParametrizedTestSuite("there_are_no_properties_for_this_name");
        
        assertEquals("Number of test cases", 2, suite.testCount());
        assertEquals("Expected the suite to contain only warnings",
                     ParametrizedTestSuite.Warning.class.getName(),
                     suite.testAt(0).getClass().getName());
        assertEquals("Expected the suite to contain only warnings",
                     ParametrizedTestSuite.Warning.class.getName(),
                     suite.testAt(1).getClass().getName());
        
        TestResult result = new TestResult();
        suite.testAt(0).run(result);
        suite.testAt(1).run(result);
        Enumeration failures = result.failures();
        assertEquals("warning message #1",
                     "Cannot load test data: Resource 'there_are_no_properties_for_this_name.properties' not found!",
                     ((TestFailure)failures.nextElement()).thrownException().getMessage());
        assertEquals("warning message #2",
                     "Test data definition for test suite 'there_are_no_properties_for_this_name' is empty!",
                     ((TestFailure)failures.nextElement()).thrownException().getMessage());
    }
    
    /** Test the {@link ParametrizedTestSuite#ParametrizedTestSuite(Class)} constructor with a class
     *  that has a matching test data properties file.
     */
    public void testParametrizedTestSuite_Class() throws Throwable {
        ParametrizedTestSuite suite = 
            new ParametrizedTestSuite(ParametrizationTestDummy.class);
        assertTestSuiteOK(suite);
    }
    
    /** Test of addTest method, of class de.tivano.junit.ParametrizedTestSuite. */
    public void testAddTest() throws Throwable {
        ParametrizedTestSuite suite = 
            new ParametrizedTestSuite("Test1", testProperties);
        suite.addTest(new ParametrizationTestDummy("testFailAlways"));
        suite.addTest(new ParametrizationTestDummy("testStandardData"));
        suite.addTest(new ParametrizationTestDummy("testSetterMethodAndPublicField"));
        suite.addTest(new ParametrizationTestDummy("testLeaveMeAlone"));
        assertEquals("Number of tests:", 4, suite.testCount());
        assertEquals("Total number of test cases:", 4, suite.countTestCases());
        for (int i=0; i<suite.testCount(); i++)
            assertWrappedTestOK(suite.testAt(i), "Test1");
    }
    
    /** Test the correct behaviour for ambiguous setter methods. */
    public void testAmbiguousSetterMethod() {
        Properties prop = new Properties();
        prop.setProperty("Test1.foo", "bar");
        ParametrizedTestSuite suite = 
            new ParametrizedTestSuite("Test1", prop);
        Test test = new Test() {
            public void run(TestResult result) {}
            public int countTestCases() { return 1; }
            public void setFoo(String value) {}
            public void setFoo(int value) {}
            
        };
        
        suite.addTest(test);
        Test wrapper = suite.testAt(0);
        TestResult result = new TestResult();
        wrapper.run(result);
        assertEquals("Number of failure in the test result:", 1, result.failureCount());
        TestFailure failure = (TestFailure)result.failures().nextElement();
        assertEquals("Error message:", 
                     "Cannot set value for attribute 'foo': More than one setter method with compatible parameter types found.", 
                     failure.thrownException().getMessage());
    }
    
    /** Test the correct behaviour for missing fields or setter methods. */
    public void testMissingProperty() {
        Properties prop = new Properties();
        prop.setProperty("Test1.foo", "bar");
        ParametrizedTestSuite suite = 
            new ParametrizedTestSuite("Test1", prop);
        Test test = new Test() {
            public void run(TestResult result) {}
            public int countTestCases() { return 1; }
        };
        
        suite.addTest(test);
        Test wrapper = suite.testAt(0);
        TestResult result = new TestResult();
        wrapper.run(result);
        assertEquals("Number of failure in the test result:", 1, result.failureCount());
        TestFailure failure = (TestFailure)result.failures().nextElement();
        assertEquals("Error message:", 
                     "Cannot set value for attribute 'foo': No setter method or public member field found with that name.", 
                     failure.thrownException().getMessage());
    }
    
    /** Test of addTestSuite method, of class de.tivano.junit.ParametrizedTestSuite. */
    public void testAddTestSuite() throws Throwable {
        ParametrizedTestSuite suite = 
            new ParametrizedTestSuite(testProperties);
        suite.addTestSuite(ParametrizationTestDummy.class);
        assertTestSuiteOK(suite);
    }
    
    
    /** Checks if <code>test</code> is a parmetization decorator, and it is parametrizing
     *  its data according to {@link #testProperties} and <code>key</code>
     * The wrapped test is assumed to be an instance of {@link ParametrizationTestDummy}
     */
    private void assertWrappedTestOK(Test wrapper, String key) throws Throwable {
        assertTrue("Test wrapper is not an instance of ParametrizedTestSuite.ParametrizedTestWrapper", 
                   wrapper instanceof ParametrizedTestSuite.ParametrizedTestWrapper);
        Test wrappedTest = ((ParametrizedTestSuite.ParametrizedTestWrapper)wrapper).getTest();
        assertTrue("Actual test is not an instance of ParametrizationTestDummy", 
                    wrappedTest instanceof ParametrizationTestDummy);
        // Test the parametrization. Because parametrizetion occurs just before
        // running the wrapped test, we need to actually run the wrapper here.
        // The ParametrizationTestDummy does the actual work of validating the
        // parametrization.
        // Before starting the test, however, we need to tell the wrapped test
        // what parametrization data it has to expect...
        ((ParametrizationTestDummy)wrappedTest).setExpectedData(key, testProperties);
        TestResult result = new TestResult();
        wrapper.run(result);
        // If the result had any errors, report this immediately. Ignore 
        // all but the first error on the assumption that they will eventually
        // be reported on another test run once the cause for the first error has been
        // fixed...
        if (result.errorCount() != 0) {
            throw ((TestFailure)result.errors().nextElement()).thrownException();
        }
        if ("testFailAlways".equals(((TestCase)wrappedTest).getName())) {
        } else {
            // If there was a failure, report it. Ignore all but the first failure,
            // they will be reported on subsequent tests once the cause for the 
            // first one has been fixed...
            if (result.failureCount() != 0) {
                throw ((TestFailure)result.failures().nextElement()).thrownException();
            }
        }
    }
    
    /** Checks if <code>suite</code> has the correct structure, and that the contained
     * tests are parametrized in the right way.
     * The structure is expected to match the definition given in the 
     * {@link ParametrizedTestSuite} documentation, with the tested class being 
     * {@link ParametrizationTestDummy} and the associated test data properties
     * being {@link #testData}.
     */
    private void assertTestSuiteOK(TestSuite suite) throws Throwable {
        assertEquals("Number of tests in the suite:", 2, suite.testCount());
        assertEquals("Total number of test cases:", 8, suite.countTestCases());
        final String TEST_NAMES[] = { "Test1", "Test123" };
        for (int i=0; i<suite.testCount(); i++) {
            String testName = TEST_NAMES[i];
            assertEquals("Inner test suite class:", 
                          ParametrizedTestSuite.class.getName(),
                          suite.testAt(i).getClass().getName());
            TestSuite innerSuite = (TestSuite)suite.testAt(i);
            assertEquals("Inner test suite name:", testName, 
                          innerSuite.getName());
            assertEquals("Number of tests in inner suite " + innerSuite.getName() + ":",
                         4, innerSuite.testCount());
            assertEquals("Total number of tests cases inner suite " + innerSuite.getName() + ":",
                         4, innerSuite.countTestCases());
            Set testCaseNames = new HashSet();
            for (int j=0; j<innerSuite.testCount(); j++) {
                assertWrappedTestOK(innerSuite.testAt(j), testName);
                String currentTest = 
                    ((TestCase)((ParametrizedTestSuite.ParametrizedTestWrapper)innerSuite.testAt(j)).getTest()).getName();
                testCaseNames.add(currentTest);
            }
            assertTrue("Missing the ParametrizationTestDummy.testFailAlways() test", 
                       testCaseNames.contains("testFailAlways"));
            assertTrue("Missing the ParametrizationTestDummy.testStandardData() test", 
                       testCaseNames.contains("testStandardData"));
            assertTrue("Missing the ParametrizationTestDummy.testSetterMethodAndPublicField() test", 
                       testCaseNames.contains("testSetterMethodAndPublicField"));
            assertTrue("Missing the ParametrizationTestDummy.testLeaveMeAlone() test", 
                       testCaseNames.contains("testLeaveMeAlone"));
        }
    }
}
