/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site (http://www.enhydra.org/).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *  @author     simon.tuffs@lutris.com
 *  @author     christos@dynamic.net.au
 *  @author     kevin@bluedog.com.au
 *  @author     ericb@interactiveportal.com
 *  @author     Christian Cryder [christianc@granitepeaks.com]
 *  @author     Stephen Peterson [stephen_peterson@agilent.com]
 *  @author     Jacob Kjome [hoju@visi.com]
 *  @author     David Li [taweili@yahoo.com]
 *  @author     Stefan Armbruster [sarmbruster@web.de]
 *
 * $Id: Xmlc.java,v 1.4 2003/12/12 05:54:02 jkjome Exp $
 */
package org.enhydra.xml.xmlc.taskdef;

import org.apache.tools.ant.*;
import org.apache.tools.ant.types.*;
import org.apache.tools.ant.types.*;
import java.lang.reflect.*;
import java.io.File;
import java.util.*;
import java.net.URL;
import java.net.MalformedURLException;


/**
 * ANT task to invoke the XMLC compiler to generate Java code.
 *
 *  <p>Implements an Ant taskdef to perform XMLC compilations. Please refer
 *  to the <a href="http://ant.apache.org/">Ant website</a>
 *  for more information. XMLCTask extends MatchingTask which means that
 *  it has the ability to include or exclude files based on pattern matching.
 *  The supported tags are as follows:</p>
 *
 *  <table border="1" cellspacing="3" cellpadding="3" width="100%">
 *    <tr>
 *      <th>tag</th>
 *      <th>required</th>
 *      <th>default value</th>
 *      <th>description</th>
 *    </tr>
 *    <tr>
 *      <td>srcdir</td>
 *      <td>yes</td>
 *      <td>&nbsp;</td>
 *      <td>The source directory root</td>
 *    </tr>
 *    <tr>
 *      <td>sourceout</td>
 *      <td>no</td>
 *      <td>value of srcdir</td>
 *      <td>The destination directory into which the generated source *ML.java files are
 *          written, if keep=&quot;yes&quot;. If sourceout is not specified it is set to the value
 *          of the srcdir. This will put the *ML.java files in the same location as the .*ml files.
 *      </td>
 *    </tr>
 *    <tr>
 *      <td>destdir</td>
 *      <td>no</td>
 *      <td>&nbsp;</td>
 *      <td>The destination directory root. Only really useful if you want XMLC to
 *          do the compilation, so we can check dependencies.
 *      </td>
 *    </tr>
 *    <tr>
 *      <td>packagedir</td>
 *      <td>no</td>
 *      <td>relative path from srcdir</td>
 *      <td>Package prefix for generated classes, expressed as a directory path (relative to srcdir).  Eg...  my/package.
 *          If not specified, the relative path from srcdir becomes the package.
 *      </td>
 *    </tr>
 *    <tr>
 *      <td>packagename</td>
 *      <td>no</td>
 *      <td>blank package</td>
 *      <td>Package prefix for generated classes, while preserving any source file directory paths.
 *          For example, given &lt;xmlc srcdir=&quot;foo/resources&quot; sourceout=&quot;foo/source&quot; packagename=&quot;a.b&quot; /&gt;, packagename=&quot;a.b&quot;,
 *          markup file located at &quot;foo/resources/c/d/Foo.ml&quot; will be generated to &quot;foo/source/a/b/c/d/FooML.java&quot; and will start with a &quot;package a.b.c.d;&quot; statement.
 *      </td>
 *    </tr>
 *    <tr>
 *      <td>markup</td>
 *      <td>no</td>
 *      <td>HTML</td>
 *      <td>The kind of markup language being processed.  Choices currently include
 *      HTML, WML, XHTML, CHTML and VoiceXML.  The generated class is suffixed with
 *      this value</td>
 *    </tr>
 *    <tr>
 *      <td>verbose</td>
 *      <td>no</td>
 *      <td>no</td>
 *      <td>Displays the XMLC command and all of the arguments</td>
 *    </tr>
 *    <tr>
 *      <td>compile</td>
 *      <td>no</td>
 *      <td>no</td>
 *      <td>Have XMLC compile the resulting java files? Note: The default is to not
 *          invoke the Java compiler through XMLC, since this is really slow. Any
 *          java files created are better handled by the ANT <javac> rule.
 *      </td>
 *    </tr>
 *    <tr>
 *      <td>force</td>
 *      <td>no</td>
 *      <td>no</td>
 *      <td>Force XMLC to always re-process the ML code whether it has changed or not</td>
 *    </tr>
 *    <tr>
 *      <td>keep</td>
 *      <td>no</td>
 *      <td>yes</td>
 *      <td>Keep the Java files generated by XMLC? To not keep them, you must allow XMLC to
 *          compile them first. However, since this is slower, the default is to keep them
 *          and let ANT compile the files using a seperate task.
 *      </td>
 *    </tr>
 *    <tr>
 *      <td>performExec</td>
 *      <td>no</td>
 *      <td>no</td>
 *      <td>XMLC is invoked by performing a Runtime.exec()<br><i>(kept for historical reasons by the author)</i></td>
 *    </tr>
 *    <tr>
 *      <td>options</td>
 *      <td>no</td>
 *      <td>&nbsp;</td>
 *      <td>File name of XMLC options or a glob style pattern.</td>
 *    </tr>
 *    <tr>
 *      <td>&lt;arg&gt;</td>
 *      <td>no</td>
 *      <td>&nbsp;</td>
 *      <td>Nested element representing additional command line arguments to be passed to XMLC.  See the XMLC documentation
 *          for all the possible commands. Note: these are applied after the other XMLC arguments derived from the options listed above.
 *          <p>For example, to specify an URL mapping to XMLC:
 *          <code>
 *            <pre>
 *            &lt;xmlc source="." includes="*.html"&gt;
 *              &lt;arg value="-urlmapping" /&gt;
 *              &lt;arg value="Edit.html" /&gt;
 *              &lt;arg value="To.po" /&gt;
 *            &lt;/xmlc&gt;
 *            </pre>
 *          </code>
 *          <b>Caution:</b> Do not be tempted to surround the URL's with single-quotes, as is common
 *          when calling XMLC from a shell command line, for example
 *            <pre>
 *            &lt;xmlc source="." includes="*.html"&gt;
 *              &lt;arg line="-urlmapping 'Edit.html' 'To.po'" /&gt;
 *            &lt;/xmlc&gt;
 *            </pre>
 *          The single-quotes will be taken as part of the URL to replace, and this is probably not what
 *          you intend
 *      </td>
 *    </tr>
 *
 *  </table>
 * <p>
 *
 * @author     simon.tuffs@lutris.com
 * @author     christos@dynamic.net.au
 * @author     kevin@bluedog.com.au
 * @author     ericb@interactiveportal.com
 * @author     Christian Cryder [christianc@granitepeaks.com]
 * @author     Stephen Peterson [stephen_peterson@agilent.com]
 * @author     Jacob Kjome [hoju@visi.com]
 * @author     David Li [taweili@yahoo.com]
 * @author     Stefan Armbruster [sarmbruster@web.de]
 * @version    %I%, %G%
 * @since      2.2
 */
public class Xmlc extends org.apache.tools.ant.taskdefs.MatchingTask {

    protected static String logHdr = "     [xmlc] ";
    protected File srcDir, destDir, srcOutDir;
    protected String classPath, packageDir, packageName, srcDirName, srcOutDirName;
    protected String options = null;
    protected String xmlcArgs = null;
    protected String performExec = null;
    protected boolean verbose, forceBuild, compile;
    protected boolean keep = true;
    protected int logLevel = project.MSG_VERBOSE;
    protected int upToDate = 0;
    // The following string is used as a suffix to the class name when generating
    // code.  This will vary, for HTML it defaults to "HTML", but other choices such
    // as "XML", "WML" etc would be appropriate.  This is set with the markup option.
    protected String ML = "HTML";
    protected Commandline cmdl = new Commandline();

    /**
     * Set the source directory containing the HTML files.
     *
     * @param srcdirname The name of the directory containing the HTML files
     */
    public void setSrcdir(String srcdirname) {
        srcDirName = srcdirname;
        srcDir = project.resolveFile(srcDirName);
    }

    /**
     * Set the destination directory into which the generated classes are
     * written.
     *
     * @param destDirName The destination directory into which the generated classes are written.
     */
    public void setDestdir(String destDirName) {
        destDir = project.resolveFile(destDirName);
    }

    /**
     * Allow general XMLC options to be passed through unchanged.  Appended to the
     * options lists.
     *
     * @param args The XMLC options to be passed through.
     * @deprecated Use &lt;arg&gt; tags for now, this will eventually be removed.
     */
    public void setArgs(String args) {
        this.xmlcArgs = args;
        project.log(logHdr+"DEPRECATED, The 'args' attribute.  Use <arg> tags instead.", project.MSG_WARN);
    }

    /**
     * Set the classpath for the XMLC compiler.
     *
     * @param classpath The classpath for XMLC to use.
     * @deprecated Use nested <arg> elements or "args" option instead.
     */
    public void setClasspath(String classpath) {
        this.classPath = classpath;
        project.log(logHdr+"DEPRECATED, The 'classpath' attribute.  Use <arg> tags instead.", project.MSG_WARN);
    }

    /**
     * Set the destination directory into which the generated source .java files are
     * written, if -keep is used.
     *
     * @param sourceout The destination directory into which the generated sources are written.
     *
     * Note: Currently -keep is used by default since the .java files are best
     * compiled by the built-in Ant javac command as a separate rule.
     */
    public void setSourceout(String sourceout) {
        this.srcOutDirName = sourceout;
        this.srcOutDir = project.resolveFile(sourceout);
    }

    /**
     * Force XMLC to always re-process the ML code (or not)
     *
     * @param force whether to force re-processing of markup or not
     */
    public void setForce(BooleanAttribute force) {
        this.forceBuild = force.booleanValue();
    }

    /**
     * Run XMLC in verbose mode (-verbose flag)
     *
     * @param verbose Set "true" for -verbose, anything else (e.g. "false") for
     * quiet mode.
     */
    public void setVerbose(BooleanAttribute verbose) {
        this.verbose = verbose.booleanValue();
        // Set the project log level to INFO if the "verbose" argument is used in the Ant invocation
        // of this taskdef.
        if (this.verbose) {
            logLevel = project.MSG_INFO;
        } else {
            logLevel = project.MSG_VERBOSE;
        }
    }

    /**
     * Set whether or not to keep the generated Java file.
     *
     * @param keep Set to "true" or "false"
     */
    public void setKeep(BooleanAttribute keep) {
        this.keep = keep.booleanValue();
    }

    /**
     * Set whether or not to have XMLC compile the generated Java file.
     *
     * @param compile Set to "true" or "false"
     */
    public void setCompile(BooleanAttribute compile) {
        this.compile = compile.booleanValue();
    }

    /**
     * Set to a specific options file to add parameters to the XMLC command. Or,
     * use a glob style pattern to base the options file name off the name of each
     * source file to be processed (e.g. *.xmlc).
     *
     *  @param options Set to file name or glob pattern
     */
    public void setOptions(String options) {
        this.options = options;
    }

    /**
     * Set the Package prefix for generated classes.  Expressed as a directory path.
     *
     * <p>Note: packagedir is used to set the package name of the generated class, <b>and</b>
     * to specify the directory where the file will be generated (relative to sourceout).
     * As such, any directory prefix on a source file is stripped before the packagedir
     * is applied.  This means that all files from the source tree will end up in the
     * same directory, which may not be what you intend, if you are using Xmlc to
     * process files in a directory hierarchy.</p>
     *
     * @param packagedir Expressed as a directory path, either '/' or '\' are accepted.
     * @see #setPackagename(String)
     */
    public void setPackagedir(String packagedir) {
        packageDir = packagedir;
    }

    /**
     * Set the package prefix for generated classes, while preserving any source
     * file directory paths.  This is useful if you want to take source (markup
     * language) files from one tree, but generate the Java code into another
     * tree, preserving relative paths.
     * <p>
     * For example, suppose that you have .html files under a directory called
     * "foo/resources", and you want to generate Java code under "foo/source",
     * but have the package name be "a.b". The following Xmlc taskdef will
     * achieve this:
     * <p>
     * &lt;xmlc srcdir=&quot;foo/resources&quot; sourceout=&quot;foo/source&quot; packagename=&quot;a.b&quot; /&gt;
     * <p>
     * A markup file located at "foo/resources/c/d/Foo.ml" will be generated to
     * "foo/source/a/b/c/d/FooML.java" and will start with a "package a.b.c.d"
     * statement.
     *
     * @param packagename Expressed as a package path, [package].[package]
     * @see #setPackagedir(String)
     */
    public void setPackagename(String packagename) {
        this.packageName = packagename;
    }

    /**
     * The kind of markup language being processed.  Choices currently include
     * HTML, WML, XHTML, CHTML and VoiceXML.  The generated class is suffixed with
     * this value.
     * 
     * @param markup The type of markup language (controls only the file-name suffix,
     * the actual markup is controlled through the XMLC -dom option which is accessible
     * through the "args" option or &lt;arg&gt; tag.
     */
    public void setMarkup(String markup) {
        this.ML = markup;
    }

    /**
     *  Set whether or not to invoke XMLC by using Runtime.exec() to
     *  create a seperate JVM. This capability was originally written to get
     *  around a problem with name-space clashes between packages. Set this
     *  to the command to execute.
     *
     *  @param performExec Set to "true" or "false"
     */
    public void setPerformexec(String performExec) {
        this.performExec = performExec;
    }

    /**
     * Check if the given target file is newer than all of the files
     * @param traget the file to be checked
     * @param files the files
     */
    private boolean isNewerThanFiles(File target, File[] files) {
        boolean newer = true;
        for (int i = 0 ; i < files.length ; i++) {
            if (files[i].lastModified() > target.lastModified()) {
                newer = false;
                project.log(logHdr+"    " + files[i].getPath() + " is newer then " + target.getPath(), logLevel);
                break;
            }
        }
        return newer;
    }

    /**
     *  Required by ant framework - implements execute() in Task.
     *  Called by the project to let the task do its work.
     *
     *  @exception BuildException All exceptions are thrown as BuildException.
     *
     * Note: most of the logic here was cloned from the Javac command. Unfortunately
     * this was the only way to 'reuse' it.
     */
    public void execute() throws org.apache.tools.ant.BuildException {
        project.log(logHdr+"Xmlc srcdir=" + srcDir, logLevel);
        project.log(logHdr+"     destdir=" + destDir, logLevel);
        project.log(logHdr+"     sourceout=" + srcOutDir, logLevel);
        project.log(logHdr+"     packagedir=" + packageDir, logLevel);
        project.log(logHdr+"     packagename=" + packageName, logLevel);
        // first off, make sure that we've got a srcdir

        if (srcDir == null) {
            throw new BuildException("srcdir attribute must be set!");
        }
        if (!srcDir.exists()) {
            throw new BuildException("srcdir does not exist!");
        }
        if (!compile && !keep) {
            throw new BuildException("If you do not keep the source, you must have XMLC compile them!");
        }

        XmlcUtils xmlcUtils = XmlcUtils.create();

        // scan source and dest dirs to build up both copy lists and
        // compile lists
        DirectoryScanner ds = this.getDirectoryScanner(srcDir);
        String[] files = ds.getIncludedFiles();

        project.log(logHdr+"XMLC processing " + files.length + " files: ", logLevel);
        for (int i=0; i<files.length; i++) {
            String file = files[i];
            try {
                String fileName = xmlcUtils.buildFullBaseFileName(packageName, packageDir, file);

                String className = xmlcUtils.buildClassName(fileName, ML);
                String javaName = className.replace('.', '/');

                // If srcOutDir is not specified, set it to srcDir.  This will put the
                // ML.java files in the same location as the .ml files, which is
                // desirable.
                if (srcOutDir == null) {
                    srcOutDir = srcDir;
                    srcOutDirName = srcDirName;
                }

                File mlFile = new File(srcDir, file);
                //csc        File javaFile = new File(srcOutDir, fileName + ML + ".java");
                File javaFile = new File(srcOutDir, javaName + ".java");  //csc
                File classFile = new File((destDir != null) ? destDir : srcOutDir, fileName + ".class");

                java.util.ArrayList args = new java.util.ArrayList();

                // Quick test: if the .java file is newer than the .ml file, don't need
                // to run XMLC.  This is a rats-nest of logic, but hard to simplify.
                File target = null;

                // The 'options' arg must  precede the name of the file to be compiled
                // Should really have -compile and -nokeep options since -nocompile
                // and -keep are the default.

                // sam, 2003-01-26
                // add recursive search
                ArrayList al = new ArrayList();
                File f = javaFile;
                while ((f = f.getParentFile()) != null) {
                    al.add(f);
                }
                f = mlFile;
                while ((f = f.getParentFile()) != null) {
                    al.add(f);
                }
                File[] possibleOptionFilesDirs = new File[al.size()];
                for (int ii=0; ii<al.size(); ii++) {
                    possibleOptionFilesDirs[al.size() - ii - 1] = (File)al.get(ii);
                    project.log("possibleOptionFilesDirs :"  + possibleOptionFilesDirs[al.size() - ii - 1].getAbsolutePath(), logLevel);
                }
                //File[] possibleOptionFilesDirs = { javaFile.getParentFile(), mlFile.getParentFile()};
                // END sam, 2003-01-26

                String[] possibleOptionFileNames = { fileName, options};
                String[] foundOptionsFiles =
                    xmlcUtils.getOptionFiles(possibleOptionFilesDirs, possibleOptionFileNames);

                File[] foundOptions = new File[foundOptionsFiles.length];
                for (int j = 0 ; j < foundOptions.length ; j++) {
                    foundOptions[j] = new File(foundOptionsFiles[j]);
                }

                /*
                 * Since we can't update the last modified time of the
                 * ML, an option checkfile is created along with the
                 * XMLC output files and this files is checked against
                 * the option files.
                 */
                File optionCheck = null;
                if (keep) {
                    optionCheck = new File(javaFile.getPath() + ".optionCheck");
                } else {
                    optionCheck = new File(classFile.getPath() + ".optionCheck");
                }

                project.log(logHdr+"    option check to created " + optionCheck.getPath(), logLevel);
                if (keep) {
                    if (!javaFile.exists()) {
                        target = javaFile;
                        // If we are going to keep the generated file but don't have one, then better create it
                        project.log(logHdr+"    " + javaFile + " does not exist, invoking XMLC", logLevel);
                    } else if (compile) {
                        target = classFile;
                        // If we are going to keep the generated file and compile it, then we have two dependencies to verify
                        if (mlFile.lastModified() <= classFile.lastModified() && mlFile.lastModified() <= javaFile.lastModified()) {
                            if (optionCheck.exists() && isNewerThanFiles(optionCheck, foundOptions)) {
                                if (!forceBuild) {
                                    project.log(logHdr+"    " + classFile.getName() + " and " + javaFile.getName() + " are newer than " + mlFile.getName(), logLevel);
                                    upToDate++;
                                    continue;
                                }
                            } else {
                                project.log(logHdr+"    " + "One of the option files are newer than " + mlFile.getName() + " invoking XMLC" , logLevel);
                            }
                        } else {
                            project.log(logHdr+"    " + classFile.getName() + " or " + javaFile.getName() + " is older than " + mlFile.getName() + " invoking XMLC", logLevel);
                        }
                    } else {
                        // If we are going to keep the generated file but not compile it, then we only care if the java
                        // file is out of date. We'll let someone else check if the class file is out of date.
                        target = javaFile;
                        if (mlFile.lastModified() <= javaFile.lastModified()) {
                            if (optionCheck.exists() && isNewerThanFiles(optionCheck, foundOptions)) {
                                if (!forceBuild) {
                                    project.log(logHdr+"    " + javaFile.getName() + " is newer than " + mlFile.getName(), logLevel);
                                    upToDate++;
                                    continue;
                                }
                            } else {
                                project.log(logHdr+"    " + "One of the option files are newer than " + mlFile.getName() + " invoking XMLC" , logLevel);
                            }
                        } else {
                            project.log(logHdr+"    " + javaFile.getName() + " is older than " + mlFile.getName() + " invoking XMLC", logLevel);
                            target = javaFile;
                        }
                    }
                } else {
                    // Not "keep" so target must be classFile.
                    target = classFile;
                    if (mlFile.lastModified() <= classFile.lastModified()) {
                        if (optionCheck.exists() && isNewerThanFiles(optionCheck, foundOptions)) {
                            if (!forceBuild) {
                                // If we are not going to keep the generated file then we must compile it
                                project.log(logHdr+"    " + classFile.getName() + " is newer than " + mlFile.getName(), logLevel);
                                upToDate++;
                                continue;
                            }
                        } else {
                            project.log(logHdr+"    " + "One of the option files are newer than " + mlFile.getName() + " invoking XMLC" , logLevel);
                        }
                    } else {
                        project.log(logHdr+"    " + classFile.getName() + " is older than " + mlFile.getName() + " invoking XMLC", logLevel);
                    }
                }

                project.log(logHdr+"    " + mlFile.getName() + " --> " + target.getName(), logLevel);
                if (classPath != null) {
                    args.add("-classpath");
                    args.add(classPath);
                }
                // Don't do the Java compilation here, it's much faster to do it with
                // the Javac task which invokes the Javac compiler without forking a
                // new process.
                if (!compile) args.add("-nocompile");
                if (keep) args.add("-keep");
                if (verbose) args.add("-verbose");

                args.add("-class");
                args.add(className);

                args.add("-sourceout");
                args.add(srcOutDir.toString());

                if (xmlcArgs != null) {
                    // We have to tokenize the arguments.  Assumpions: URL's will not contain
                    // whitespace, and if they did, they should use %20 or &space entities.
                    //
                    java.util.StringTokenizer tok = new java.util.StringTokenizer(xmlcArgs, " ");
                    while (tok.hasMoreElements()) {
                        args.add(tok.nextToken());
                    }
                }

                String[] additionalArgs = cmdl.getArguments();
                for (int idx=0; idx < additionalArgs.length; idx++) {
                    args.add(additionalArgs[idx]);
                }

                // Option files to be added right before the ML file
                if (foundOptionsFiles.length > 0) {
                    for (int optionsIdx = 0; optionsIdx < foundOptionsFiles.length; optionsIdx++) {
                        args.add(foundOptionsFiles[optionsIdx]);
                        project.log(logHdr+"    Using options file " + foundOptionsFiles[optionsIdx], logLevel);
                    }
                } else {
                    project.log(logHdr+"No options file could not be located for " + file +
                        " in paths " + javaFile.getParentFile() + ", " +
                        mlFile.getParentFile(), project.MSG_WARN);
                }
                // *** WARNING DO NOT SEPARATE THE FOLLOWING CODE FROM THE BLOCK ABOVE! ***
                args.add(new File(srcDir, file).getAbsolutePath());

                project.log(logHdr+"args==" + args, logLevel);

                project.log(logHdr+"Invoke XMLC on " + mlFile.getPath(), project.MSG_INFO);
                if (performExec != null) {
                    args.add(0, performExec);
                    String [] _args = (String[])args.toArray(new String[args.size()]);
                    Runtime runtime = Runtime.getRuntime();
                    Process process = runtime.exec(_args);
                    int status = process.waitFor();
                    if (status != 0) {
                        throw new BuildException("exit status of exec = "+status + toString(_args));
                    }
                } else {
                    String [] _args = (String[])args.toArray(new String[args.size()]);

                    // Execute XMLC directly.  This requires that enhydra.jar, or an equivalent
                    // path containing the needed DOM code is on the CLASSPATH of ANT.
                    String xmlcClassName = "org.enhydra.xml.xmlc.commands.xmlc.XMLC";

                    Method m = null;
                    Class  c = null;
                    try {
                        c = Class.forName(xmlcClassName, true, getClass().getClassLoader());
                        m = c.getMethod("main", new Class[] { String[].class});
                    } catch (Exception e) {
                        // Reflection errors.  Class or method does not exist!
                        e.printStackTrace();
                        throw new BuildException(e);
                    }

                    if (m != null) {
                        m.invoke(null,new Object[] { _args});
                    }
                }

                /*
                 * Create the option file
                 */
                if (optionCheck.exists()) {
                    optionCheck.delete();
                }
                optionCheck.createNewFile();
            } catch (InvocationTargetException ite) {
                ite.printStackTrace();
                throw new BuildException(ite);
            } catch (Exception x) {
                x.printStackTrace();
                throw new BuildException(x);
            }
        }

        if (upToDate > 0) {
            project.log(logHdr+"XMLC " + upToDate + " files were up-to-date, consider using force=\"true\"", logLevel);
        }
    }

    /**
     *  Simple utility to concatenate an array of Strings into one single
     *  String.
     *
     *  @param      args  The array of Strings
     *  @return     The resulting String
     */
    static private String toString(String[] args) {
        StringBuffer  buffer = new StringBuffer();
        for (int i=0; i<args.length; i++) buffer.append(args[i]).append(" ");
        return buffer.toString();
    }

    /**
     *  Set nested arguments to the XMLC command to be executed.
     *
     */
    public Commandline.Argument createArg() {
        return cmdl.createArgument();
    }

    public static class BooleanAttribute extends EnumeratedAttribute {
        public String[] getValues() {
            return new String[] {"yes", "no", "on", "off", "true", "false"};
        }

        public boolean booleanValue() {
            return (this.getValue().equals("yes") || this.getValue().equals("on") || this.getValue().equals("true"));
        }
    }

}
