/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site (http://www.enhydra.org/).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 * $Id: XmlcUtils.java,v 1.1 2003/05/06 18:32:39 jkjome Exp $
 */
package org.enhydra.xml.xmlc.taskdef;

import java.io.File;


/**
 * Abstract base xmlc utilities class
 *
 * @author Robert Leftwich
 * @version 1.0
 */
public abstract class XmlcUtils {

    /**
     * The default extension for an XMLC options file.
     */
    public static final String DEFAULT_OPTIONS_FILETYPE = ".xmlc";

    /**
     * The name of the default XMLC options file.
     */
    public static final String DEFAULT_OPTIONS_FILENAME = "options" + DEFAULT_OPTIONS_FILETYPE;

    /**
     * The system property holding the classname of the class to
     * instantiate when XmlcUtils.create() is called.
     */
    public static final String DEFAULT_UTILS_CLASS_KEY = "xmlcutils.classname";

    /**
     * The default classname of the class to
     * instantiate when XmlcUtils.create() is called.
     */
    public static final String DEFAULT_UTILS_CLASS_NAME = DefaultXmlcUtilsImpl.class.getName();


    /**
     * Construct default XmlcUtils (protected to force the use
     * of XmlcUtils.create() factory method).
     */
    protected XmlcUtils() {}

    //===========================================================================
    // QUERY METHODS
    //===========================================================================

    /**
     * Get a list of options files for the specified directory.
     * This method will search for the default options file name
     * (options.xmlc) in the specified directory and the user.dir
     * directory as well. It will return them in that order.
     *
     * @param theDirectory
     *               The directory to look in for the options file.
     * @return An array of options files with absolute paths.
     *         An empty array if none found.
     */
    public abstract String[] getOptionFiles(File theDirectory);

    /**
     * Get a list of options files for the specified directory,
     * having the specified name. This method will search for the
     * specified options file name and the default options file name
     * (options.xmlc) in the specified directory and the user.dir
     * directory as well. It will return them in that order.
     *
     * @param theDirectory
     *               The directory to look in for the options file.
     * @param theOptionsFileName
     *               The name of the options file to look for.
     * @return An array of options files with absolute paths.
     *         An empty array if none found.
     */
    public abstract String[] getOptionFiles(File theDirectory, String theOptionsFileName);

    /**
     * Get a list of options files for the specified directory,
     * having the specified names. This method will search for the
     * specified options file names and the default options file name
     * (options.xmlc) in the specified directory and the user.dir
     * directory as well. It will return them in that order.
     *
     * @param theDirectory
     *               The directory to look in for the options file.
     * @param theOptionsFileNames
     *               The list of names of the options file to look for.
     * @return An array of options files with absolute paths.
     *         An empty array if none found.
     */
    public abstract String[] getOptionFiles(File theDirectory, String[] theOptionsFileNames);

    /**
     * Get a list of options files for the specified directories,
     * having the specified name. This method will search for the
     * specified options file name and the default options file name
     * (options.xmlc) in the specified directories and the user.dir
     * directory as well. It will return them in that order.
     *
     * @param theDirectories
     *               The list of directories to look in for the options files.
     * @param theOptionsFileName
     *               The name of the options file to look for.
     * @return An array of options files with absolute paths.
     *         An empty array if none found.
     */
    public abstract String[] getOptionFiles(File[] theDirectories, String theOptionsFileName);

    /**
     * Get a list of options files for the specified directories,
     * having the specified names. This method will search for the
     * specified options file names and the default options file name
     * (options.xmlc) in the specified directories and the user.dir
     * directory as well. It will return them in that order.
     *
     * @param theDirectories
     *               The list of directories to look in for the options files.
     * @param theOptionsFileNames
     *               The list of names of the options file to look for.
     * @return An array of options files with absolute paths.
     *         An empty array if none found.
     */
    public abstract String[] getOptionFiles(File[] theDirectories, String[] theOptionsFileNames);

    /**
     * Build a full base file name (i.e. with no file type) out of
     * the specified components.
     *
     * Note that any directory prefix on the theBaseFileName file is stripped before
     * thePackageDir is applied. So that the following call :
     *
     *     buildFullBaseFileName("a.b.c", "foo/resources", "config/test.html");
     *
     * will return a string of "a/b/c/foo/resources/test"
     *
     * and this call :
     *
     *     buildFullBaseFileName("a.b.c", "", "config/test.html");
     *
     * will return a string of "a/b/c/config/test"
     *
     * @param thePackageName
     *               The name of the package to prepend to the file name.
     *               This should be specified using the Java naming convention
     *               i.e. org.enhydra.xmlc
     * @param thePackageDir
     *               The name of the directory to prepend to the file name.
     *               This is a directory path that can be specified using either '/' or '\'
     *               i.e. foo/resources
     * @param theBaseFileName
     *               The base file name that the preceding strings are prepended to.
     * @return A string representing the full path resulting from the specified components
     *         with all separators set to '/'
     */
    public abstract String buildFullBaseFileName(String thePackageName,
                                                  String thePackageDir,
                                                  String theBaseFileName);

    /**
     * Build a Java class name out of the specified components.
     *
     * @param theFullBaseFileName
     *               The full base file name (specified with '/' separators)
     *               and no file type. e.g. "org.enhydra.test"
     * @param theModifier
     *               The file modifier to append to the class name
     *               e.g. HTML
     * @return A Java class name string
     */
    public abstract String buildClassName(String theFullBaseFileName,
                                         String theModifier);

    /**
     * Factory method for creating instances of the XmlcUtils class.
     * The actual implementation instantiated is controlled by
     * setting the system property DEFAULT_UTILS_CLASSKEY
     * (currently 'xmlcutils.classname') to the fully qualified
     * class name of the sub-class of XmlcUtils that you want to use.
     * By default (or if the value specified is invalid) this
     * method will return an instance of DefaultXmlcUtilsImpl
     *
     * @return An instance of a XmlcUtils class
     */
    public static XmlcUtils create() {
        XmlcUtils retVal = null;
        String defaultUtilsClassName = System.getProperty(DEFAULT_UTILS_CLASS_KEY, DEFAULT_UTILS_CLASS_NAME);
        try {
            Class classObj = Class.forName(defaultUtilsClassName, true, XmlcUtils.class.getClassLoader());
            if (XmlcUtils.class.isAssignableFrom(classObj)) {
                retVal = (XmlcUtils)classObj.newInstance();
            } else {
                //theirLogger.warn("A \"" + defaultUtilsClassName +
                //      "\" object is not an instance of a \"" +
                //      XmlcUtils.class.getName() + "\" object.");
                System.out.println ("WARNING: A \"" + defaultUtilsClassName +
                    "\" object is not an instance of a \""+
                    XmlcUtils.class.getName() + "\" object.");
            }
        } catch (Exception e) {
            //theirLogger.warn("Could not instantiate class [" + defaultUtilsClassName + "].", e);
            System.out.println("WARNING: Could not instantiate class [" + defaultUtilsClassName + "]."+e);
        }
        if (null == retVal) retVal = new DefaultXmlcUtilsImpl();
        return retVal;
    }

}
