/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: BaseDOMFormatter.java,v 1.5 2004/04/08 04:31:33 jkjome Exp $
 */

package org.enhydra.xml.io;

import org.enhydra.xml.dom.DOMTraversal;
import org.enhydra.xml.dom.DOMError;
import org.enhydra.xml.dom.DOMOps;
import org.enhydra.xml.dom.DOMAccess;
import org.enhydra.xml.xmlc.XMLObject;
import org.enhydra.xml.xmlc.XMLObjectLink;

import java.util.HashMap;
import java.io.Writer;
import java.io.StringWriter;
import java.io.IOException;

import org.w3c.dom.*;
import org.w3c.dom.html.HTMLDocument;


/**
 * Base class with common functionally XML and HTML DOM formatting
 * functionality.
 */
abstract class BaseDOMFormatter implements Formatter, DOMTraversal.Handler {
    // N.B. A lot of fields are made final to aid in optimization.

    /**
     * Maximum character value in an entity quick-check table.
     */
    protected static final int MAX_ENTITY_QUICK_CHECK_CHAR = 0x7f;

    /**
     * Quote character to use for attribute values.
     */
    protected static final char ATTR_QUOTE_CHAR = '"';

    /**
     * Entity reference for attribute value quote character.
     */
    protected static final String ATTR_QUOTE_CHAR_ENTITY_REF = "#34";

    /**
     * The output options.
     */
    protected final OutputOptions fOptions;

    /**
     * Output writer.
     */
    protected Writer fOut;

    /**
     * DOM traverser.
     */
    protected final DOMTraversal fTraverser;

    /**
     * Should character entity references be omitted from attribute values?
     */
    private final boolean fOmitAttributeCharEntityRefs;

    /**
     * Are we doing pre-formatting?  
     */
    protected final boolean fPreFormatMode;

    /**
     * Should pre-formatted text be written.  This is only enabled if the
     * document implements PreFormattedTextDocument and the preformatted
     * encoding matches the output encoding.  This option only controls
     * text nodes that are not children of attributes.
     */
    protected final boolean fUsePreFormattedText;

    /**
     * Should pre-formatted text be use for descendents of attribute
     * nodes.  A seperate flag is required to support the
     * omitAttributeCharEntityRefs option.
     */
    protected final boolean fUsePreFormattedAttrText;

    /**
     * Should pre-formatted Elements be written.  This is the same as
     * fUsePreFormattedText, unless a URLRewriter is present, it which
     * cases its false, since we need to check the attributes for URLs.
     */
    protected final boolean fUsePreFormattedElements;

    /**
     * Count of preformatted text nodes that were written.
     */
    private int fPreFormattedTextCount;

    /**
     * Count of text nodes that were written not using preformatted text.
     */
    private int fDynamicFormattedTextCount;

    /**
     * Count of preformatted element open tags that were written.
     */
    protected int fPreFormattedElementCount;

    /**
     * Count of element open tags that were written not using preformatted
     * text.
     */
    protected int fDynamicFormattedElementCount;

    /**
     * Are we currently outputting an attribute or its children.
     */
    private boolean fProcessingAttr;

    /**
     * Are we pretty-printing?
     */
    protected final boolean fPrettyPrinting;
    
    /**
     * Pretty-printing indent size.
     */
    private final int fIndentSize;

    /**
     * The encoding for the current document.
     */
    private final String fEncoding;

    /**
     * The maximum value of an unicode character in the document's encoding.
     */
    private final int fMaxCharacterValue;

    /**
     * The character set object for the encoding.
     */
    private final CharacterSet fCharSet;

    /**
     * Object to do URL value rewriting, or null if URL rewriting is
     * not being done.
     */
    private final URLRewriter fURLRewriter;

    /**
     * Document being formatted.
     */
    protected final Document fDocument;
    
    /**
     * DocumentType from document.
     */
    protected final DocumentType fDocType;
    
    /**
     * Public id to use for DOCTYPE, or null if none.
     */
    protected final String fPublicId;

    /**
     * System id to use for DOCTYPE, or null if none.
     */
    protected final String fSystemId;

    /**
     * If the document implements DocumentInfo, this is set.  It is
     * used by URL rewriting.
     */
    private final DocumentInfo fDocInfo;

    /**
     * Entity quick-check table, indexed by character value, indicating if the
     * character must be represented as an character entity reference.
     */
    private final boolean[] fEntityQuickCheck;

    /**
     * Platform line separator.
     */
    private static final String fNewLine;

    /**
     * Static string that is used as a quick way to get a certain indent
     * level.  Indent strings are generated by taking substrings of this
     * string.  This string is grown as needed.
     */
    private static String fIndentSource
        = "                                                               ";

    /**
     * Static initializer.
     */
    static {
        fNewLine = System.getProperty("line.separator");
        if (fNewLine == null) {
            throw new XMLIOError("System property line.separator not found");
        }
    }

    /**
     * Check if preformatted text might be used for this document.
     * 
     * @return The output options that were used for preformatting or null
     *  if preformatted text can be used for some reason.
     */
    private static OutputOptions checkUsePreformatting(Document doc,
                                                       String defaultEncoding,
                                                       CharacterSet charSet) {
        // Document must implement PreFormattedTextDocument
        if (!(doc instanceof PreFormattedTextDocument)) {
            return null;
        }
        
        // OutputOptions used for preformatting must be available
        OutputOptions pfOptions = ((PreFormattedTextDocument)doc).getPreFormatOutputOptions();
        if (pfOptions == null) {
            return null;
        }

        // Make sure that the two encodings are compatible in terms of which
        // characters are converted to character entity references.
        String preFormatEncoding = pfOptions.getMIMEEncoding();
        if (preFormatEncoding == null) {
            preFormatEncoding = defaultEncoding;
        }
        Encodings encodings = Encodings.getEncodings();
        CharacterSet preFormatCharSet
            = encodings.getCharacterSet(preFormatEncoding);
        if (!charSet.isCompatible(preFormatCharSet)) {
            return null;  // not compatible
        }
        return pfOptions;
    }

    /**
     * If preformatting is enabled, determine if text nodes should 
     * use preformatted text.
     */
    private static boolean checkUsePreformattedText(OutputOptions options,
                                                    OutputOptions pfOptions) {
        // Only pretty-printing must be the same
        return (pfOptions.getIndentSize() == options.getIndentSize())
            && (pfOptions.getPrettyPrinting() == options.getPrettyPrinting());
    }

    /**
     * If preformatting is enabled, determine if attribute value text nodes
     * should use preformatted text.
     */
    private static boolean checkUsePreformattedAttrText(OutputOptions options,
                                                        OutputOptions pfOptions) {
        return (pfOptions.getOmitAttributeCharEntityRefs() == options.getOmitAttributeCharEntityRefs());
    }

    /**
     * If preformatting is enabled, determine if preformatted element open
     * tags should be used.
     */
    private static boolean checkUsePreformattedElements(OutputOptions options,
                                                        OutputOptions pfOptions,
                                                        URLRewriter urlRewriter) {
        return (pfOptions.getDropHtmlSpanIds() == options.getDropHtmlSpanIds())
            && (pfOptions.getOmitAttributeCharEntityRefs() == options.getOmitAttributeCharEntityRefs())
            && (urlRewriter == null);
    }

    /**
     * Get the DocumentInfo object, if available.
     */
    private static DocumentInfo findDocumentInfo(Document document) {
        if (document instanceof DocumentInfo) {
            return (DocumentInfo)document;
        } else if (document instanceof XMLObjectLink) {
            // Go back to the XMLObject if possible.
            return (DocumentInfo)((XMLObjectLink)document).getXMLObject();
        } else if (document instanceof HTMLDocument) {
            return new DocumentInfo () {
		    public boolean isURLAttribute(Element element,
						  String attrName) {
			return org.enhydra.xml.xmlc.dom.HTMLDomFactoryMethods.isURLAttribute(element, attrName);
		    }

		};
        } else {
	    return null;
	}
    }

    /**
     * Get the encoding to use. If one is not explictly specified in
     * output options, see if one can be obtained from the document.
     * If that fails, use the default for this type of document.
     */
    private static String getEncoding(Document document,
                                      OutputOptions outputOptions,
                                      String defaultEncoding) {
        String encoding = outputOptions.getMIMEEncoding();
        if (encoding == null) {
            // Try getting from the document
            if (document instanceof XMLObject) {
                encoding = ((XMLObject)document).getEncoding();
            } else if (document instanceof XMLObjectLink) {
                encoding = ((XMLObjectLink)document).getXMLObject().getEncoding();
            }
            if (encoding == null) {
                // ok, must use default for XML or HTML
                encoding = defaultEncoding;
            }
        }
        return encoding;
    }

    /**
     * Constructor.
     *
     * @param node Any node of the document that this formatter will be
     *  associated with.  This can also be an XMLC Document object (XMLObject).
     * @param options The output options.
     * @param defaultEncoding The default encoding for this format.
     * @param forPreFormatting Is this going to be used for preformatting?
     * @param entityQuickCheck Document-type specific table that provides
     *  a quick check of the need to encode that character as a character
     *  entity reference.  This table MUST include the double-quote character,
     *  as it it used to quote attribute values.
     */
    protected BaseDOMFormatter(Node node,
                               OutputOptions outputOptions,
                               boolean forPreFormatting,
                               String defaultEncoding,
                               boolean[] entityQuickCheck) {
        fDocument = DOMOps.getDocument(node);
        fDocType = DOMAccess.accessDocumentType(fDocument);
        fOptions = outputOptions;
        fTraverser = DOMTraversal.getTraverser(this, 0, node);

        // Get DOCTYPE information (and overrides from OutputOptions)
        String publicId = fOptions.getPublicId();
        String systemId = fOptions.getSystemId();
        if (fDocType != null) {
            if (fDocType.getPublicId() != null) {
                publicId = fDocType.getPublicId();
            }
            if (fDocType.getSystemId() != null) {
                systemId = fDocType.getSystemId();
            }
        }
        fPublicId = publicId;
        fSystemId = systemId;

        // Initialize the encoding information.
        fEncoding = getEncoding(fDocument, outputOptions, defaultEncoding);
        Encodings encodings = Encodings.getEncodings();
        fMaxCharacterValue = encodings.getMaxCharacterValue(fEncoding);
        fCharSet = encodings.getCharacterSet(fEncoding);

        // Setup URL rewriting, which needs DocumentInfo
        fDocInfo = findDocumentInfo(fDocument);
        fURLRewriter = (fDocInfo != null) ? fOptions.getURLRewriter() : null;

        // Various options from output options.
        fIndentSize = fOptions.getIndentSize();
        fPrettyPrinting = (fIndentSize > 0) && fOptions.getPrettyPrinting();
        fOmitAttributeCharEntityRefs = fOptions.getOmitAttributeCharEntityRefs();

        // Initialize preformatted text options.  This is all rather tricky.
        OutputOptions pfOptions
            = checkUsePreformatting(fDocument, defaultEncoding, fCharSet);

        if ((pfOptions != null) && (!forPreFormatting)) {
            fUsePreFormattedText = checkUsePreformattedText(fOptions, pfOptions);
            fUsePreFormattedAttrText = checkUsePreformattedAttrText(fOptions, pfOptions);
            fUsePreFormattedElements = checkUsePreformattedElements(fOptions, pfOptions,
                                                                    fURLRewriter);
        } else {
            fUsePreFormattedText = false;
            fUsePreFormattedAttrText = false;
            fUsePreFormattedElements = false;
        }
        fPreFormatMode = forPreFormatting;

        // Entity handling
        fEntityQuickCheck = entityQuickCheck;
        if (!fOptions.getUseAposEntity())
            fEntityQuickCheck['\''] = false;

        // initialize for pre-formatting.
        if (forPreFormatting) {
            fOut = new StringWriter(4096); // bigger than default
        }
    }

    /**
     * @see Formatter#getMIMEEncoding
     */
    public final String getMIMEEncoding() {
        return fEncoding;
    }

    /**
     * @see Formatter#usedPreFormattedText
     */
    public boolean usedPreFormattedText() {
        return fUsePreFormattedText;
    }

    /**
     * Get the count of preformatted text nodes that were written.
     */
    public final int getPreFormattedTextCount() {
        return fPreFormattedTextCount;
    }

    /**
     * Get the count of text nodes that were written not using
     * preformatted text.
     */
    public final int getDynamicFormattedTextCount() {
        return fDynamicFormattedTextCount;
    }

    /**
     * @see Formatter#usedPreFormattedElements
     */
    public final boolean usedPreFormattedElements() {
        return fUsePreFormattedElements;
    }

    /**
     * Get the count of preformatted element open tags that were written.
     */
    public final int getPreFormattedElementCount() {
        return fPreFormattedElementCount;
    }

    /**
     * Get the count of element open tags that were written not using
     * preformatted text.
     */
    public final int getDynamicFormattedElementCount() {
        return fDynamicFormattedElementCount;
    }

    /**
     * Write a newline
     */
    protected final void writeln() throws IOException {
        fOut.write(fNewLine);
    }

    /**
     * Grow indentation source string to at least the specified size
     * if needed.
     */
    private static void ensureIndentSource(int numChars) {
        // Grow as needed; coded to avoid sync
        while (fIndentSource.length() < numChars) {
            fIndentSource += fIndentSource;
        }
    }

    /**
     * Print indentation to the current level.
     */
    protected final void printIndent() throws IOException {
        if (fPrettyPrinting) {
            int indent = (fTraverser.getDepth() - 1) * fIndentSize;
            ensureIndentSource(indent);
            fOut.write(fIndentSource, 0, indent);
        } // end of if ()
    }

    /**
     * Get a character entity name for a character.
     * This is the slow-path, so its ok this is an abstract method call.
     * @return The character entity name, or null if this character doesn't
     *  have one.
     */
    abstract protected String getCharacterEntity(char textChar);

    /**
     * Output a character, possibly substituting a character entity reference
     * or a numeric entity. This is the slow-path method.
     */
    private void writeCharacter(char textChar) throws IOException {
        // Determine the entity to substitute; always do quote, even if substitute disabled,
        // as this is only implemented for attribute values.
        String entity;
        if (fProcessingAttr && fOmitAttributeCharEntityRefs) {
            entity = (textChar == ATTR_QUOTE_CHAR) ? ATTR_QUOTE_CHAR_ENTITY_REF : null;
        } else {
            entity = getCharacterEntity(textChar);
        }

        // FIXME: need to review impact of isValid call on performance.
        // this use to be a simple compare.
        if (entity != null) {
            fOut.write('&');
            fOut.write(entity);
            fOut.write(';');
        } else if (!fCharSet.isValid(textChar)) {
            fOut.write("&#");
            fOut.write(Integer.toString(textChar));
            fOut.write(';');
        } else {
            fOut.write(textChar);
        }
    }

    /**
     * Write a text string, encoding document type-specific character entities.
     * This is an expensive procedure and has been carefully hand optimized.
     */
    protected final void writeText(String text) throws IOException {
        if (text == null) return;
        int len = text.length();
        char ch;

        // FIXME: here we make a nasty assumption about all characters
        // less than fMaxCharacterValue && MAX_ENTITY_QUICK_CHECK_CHAR 
        // are valid. This is probably ok.  The fMaxCharacterValue check
        // is needed for 7-bit encodings.
        for (int idx = 0; idx < len; idx++) {
            ch = text.charAt(idx);
            if ((ch <= fMaxCharacterValue)
                && (ch <= MAX_ENTITY_QUICK_CHECK_CHAR)
                && (!fEntityQuickCheck[ch])) {
                fOut.write(ch);  // Fast path.
            } else {
                writeCharacter(ch);
            }
        }
    }

    /**
     * Handler called for Entity nodes; should never be called.
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleEntity
     */
    public final void handleEntity(Entity entity) {
        throw new XMLIOError("Unexpected call to handleEntity");
    }

    /**
     * Handler called for Notation nodes; should never be called.
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleNotation
     */
    public final void handleNotation(Notation notation) {
        throw new XMLIOError("Unexpected call to handleNotation");
    }

    /**
     * Handler called for EntityReference nodes.
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleEntityReference
     */
    public final void handleEntityReference(EntityReference entityRef) throws IOException {
        fOut.write('&');
        fOut.write(entityRef.getNodeName());
        fOut.write(';');
    }

    /**
     * Handler called for Comment nodes.
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleComment
     */
    public final void handleComment(Comment comment) throws IOException {
        fOut.write("<!--");
        fOut.write(comment.getData());
        fOut.write("-->");
    }

    /**
     * Handler called for Text nodes.
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleText
     */
    public void handleText(Text text) throws IOException {
        String preformattedData = null;
        if (fProcessingAttr) {
            if (fUsePreFormattedAttrText && (text instanceof PreFormattedText)) {
                preformattedData = ((PreFormattedText)text).getPreFormattedText();
            }
        } else {
            if (fUsePreFormattedText && (text instanceof PreFormattedText)) {
                preformattedData = ((PreFormattedText)text).getPreFormattedText();
            }
        }

        if (preformattedData != null) {
            // Have preformatted data that can be used.
            fOut.write(preformattedData);
            fPreFormattedTextCount++;
        } else {
            // No preformatted text.
            writeText(text.getData());
            fDynamicFormattedTextCount++;
        }
    }

    /**
     * Write an attribute value.  Convert characters to character entity
     * references as needed.
     */
    protected final void writeAttributeValue(Attr attr) throws IOException {
        fProcessingAttr = true;

        fOut.write('=');
        fOut.write(ATTR_QUOTE_CHAR);

        // Are we doing URL rewritting and is this a URL attr?
        if ((fURLRewriter != null)
            && fDocInfo.isURLAttribute(attr.getOwnerElement(),
                                       attr.getName())) {
            String value = fURLRewriter.rewriteURL(attr.getValue());
            writeText(value);
        } else {
            fTraverser.processChildren(attr);
        }
        fOut.write(ATTR_QUOTE_CHAR);

        // Don't need to reset on error, as its reset at the beginning
        // of each write.
        fProcessingAttr = false;
    }

    /**
     * Method to write an open tag, including attributes.  Children
     * are not processed.  This is normally called by the derived class
     * handleElement method, but its hear to allow for use by preformatter.
     */
    abstract protected void writeOpenTag(Element element, 
                                         String tagName,
                                         boolean hasChildren) throws IOException;

    /**
     * Preformat an element.
     */
    private String preFormatElement(Element element) throws IOException {
        writeOpenTag(element, element.getTagName(), element.hasChildNodes());
        return ((StringWriter)fOut).getBuffer().toString();
    }

    /**
     * Recursively determine if a node is a child of an attribute.
     */
    private boolean isAttributeChild(Node node) {
        if (node == null) {
            return false;
        }
        switch (node.getNodeType()) {
        case Node.ELEMENT_NODE:
            return false;
        case Node.ATTRIBUTE_NODE:
            return true;
        default:
            return isAttributeChild(node.getParentNode());
        }
    }

    /**
     * Preformat a text node.  The preformatted text is returned even if
     * its the same.
     */
    private String preFormatText(Text text) throws IOException {
        fProcessingAttr = isAttributeChild(text);
        try {
            handleText(text);
        } finally {
            fProcessingAttr = false;
        }
        return ((StringWriter)fOut).getBuffer().toString();
    }

    /**
     * @see Formatter#preFormatNode
     */
    public final String preFormatNode(Node node) {
        fProcessingAttr = false;
        try {
            ((StringWriter)fOut).getBuffer().setLength(0); // Clear buffer
            
            switch (node.getNodeType()) {
            case Node.ELEMENT_NODE:
                return preFormatElement((Element)node);
            case Node.TEXT_NODE:
                return preFormatText((Text)node);
            }
            return null;
        } catch (IOException except) {
            // Should never happen.
            throw new XMLIOError(except);
        }
    }

    /**
     * Format a Node and children to the specified writer.
     * @see Formatter#write
     */
    public final void write(Node node,
                            Writer writer) throws IOException {
        try {
            fOut = writer;
            fProcessingAttr = false;
            fTraverser.traverse(node);
//csc_040604_1 - this really causes problems if you are using the DOMWriter repeatedly, to write chunks of the document at a
//               time, rather than the whole thing at once. By putting a writeln() here, it ends up splitting lines wherever
//               the chunk finishes, and if that's in between spans surrounding text nodes, it can adversely effect the output
//               that gets displayed. Consequently, we really need to remove this writeln. What this means is that there will
//               no longer be a final CR/LF at the end of reports, but I don't think that will actually cause any problems
//csc_040604_1            writeln();
        } catch (DOMError error) {
            // Rethrow IOExceptions
            Throwable cause = error.getCause();
            if (cause instanceof IOException) {
                throw (IOException)cause;
            } else {
                throw error;
            }
        }
    }
}
