/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DocTypeBuilder.java,v 1.2 2003/06/29 05:58:48 jkjome Exp $
 */

package org.enhydra.xml.xmlc.dom;

import org.enhydra.xml.xmlc.*;
import org.w3c.dom.*;
import java.util.HashMap;

/**
 * Class for building a DOM DocumentType. Used by XMLC parsers to build a document
 * DOM.
 * <P>
 * Since the DOM level 1 does not address storing all of the information
 * from the DOCTYPE and DTDs required by XMLC, routine are provided to
 * add this information.  Only a subset of the information is actually saved.
 * The internal subset is needed to reproduce the DOCTYPE declaration in
 * the compiled document.  The attribute declarations from the external
 * subset are needed to build the table of ID attributes used to generate
 * access methods.
 */
public class DocTypeBuilder {
    /**
     * Factory for creating the document.
     */
    private XMLCDomFactory domFactory;

    /**
     * DocumentType object, null until created.
     */
    private DocumentType docType;

    /**
     * Information needed to build DocumentType.
     */
    private String docTypeName;       // Root element.
    private String docSystemId;
    private String docPublicId;

    /*
     * Internal subset as a string.
     */
    private String internalSubsetStr;

    /**
     * Flag to indicate that getCreateDocType() has been called.  After this
     * call, the attributes of this class can't be changed.  A seperate flag
     * is used, as the docType object maybe null.
     */
    private boolean createdDocType = false;

    /**
     * Table of element to id attribute name mappings.
     */
    private HashMap elementIdAttrs = new HashMap();

    /**
     * Constructor.
     * 
     * @param domFactory Factory class for Documents.
     */
    public DocTypeBuilder(XMLCDomFactory domFactory) {
        this.domFactory = domFactory;
    }

    /**
     * Get the document type object, creating if necessary.  Once created,
     * no modifications can be made to this object.  If setDocumentTypeName()
     * has not been called, null is return.  This is the cause when a
     * document doesn't have a DTD.
     */
    public DocumentType getCreateDocType() {
        if ((docType == null) && (docTypeName != null)) {
            docType = domFactory.createDocumentType(docTypeName, docPublicId,
                                                    docSystemId, internalSubsetStr);
        }
        // This is set even if we didn't actually build a DocumentType
        createdDocType = true;
        return docType;
    }

    /**
     * Generate an error if the doc type has been created, indicating
     * a bug in the code using this class.
     */
    private void checkIfAlreadyCreated() {
        if (createdDocType) {
            throw new XMLCError("XMLC bug: attempt to add document type data after DocumentType object has been created");
        }
    }

    /**
     * Set the document type name (rootElement).
     *
     * @param name The Document type name (also root node name).
     */
    public void setDocumentTypeName(String name) {
        checkIfAlreadyCreated();
        docTypeName = name;
    }
    
    /**
     * Get the document type name (rootElement).
     */
    public String getDocumentTypeName() {
        return docTypeName;
    }

    /**
     * Set the publicId.
     *
     * @param publicId Document type public id or null if standalone.
     */
    public void setPublicId(String publicId) {
        checkIfAlreadyCreated();
        docPublicId = publicId;
    }

    /**
     * Get the publicId.
     */
    public String getPublicId() {
        return docPublicId;
    }

    /**
     * Set the systemId.
     *
     * @param systemId Document type system id or null if standalone.
     */
    public void setSystemId(String systemId) {
        checkIfAlreadyCreated();
        docSystemId = systemId;
    }

    /**
     * Get the systemId.
     */
    public String getSystemId() {
        return docSystemId;
    }

    /**
     * Define an element id attribute.
     *
     * @param elementName The name of the element.
     * @param attributeName The name of the ID attribute.
     * @param internalSubset Is this part of the internal or
     *  external subset?  Internal declarations take precedence.
     */
    public void addIdAttribute(String elementName,
                               String attributeName,
                               boolean internalSubset) {
        checkIfAlreadyCreated();

        boolean exists = elementIdAttrs.containsKey(elementName);
        if ((!exists) || (exists && internalSubset)) {
            elementIdAttrs.put(elementName, attributeName);
        }
    }

    /**
     * Get the id attribute name for an element.  XML only allows one
     * id attribute per element type.
     *
     * @param elementName The name of the element.
     * @return The attribute name or null if no ID attribute is defined.
     */
    public String getIdAttribute(String elementName) {
        return (String)elementIdAttrs.get(elementName);
    }

    /**
     * Add an EntityReference object.
     *
     * @param entity The name of the entity.
     * @param internalSubset Is this part of the internal or
     *  external subset?
     */
    public void addEntityReference(String name,
                                   boolean internalSubset) {
        checkIfAlreadyCreated();
    }

    /**
     * Add a document type declaration.
     * 
     * @param name The element name.
     * @param contentSpec The content specification.
     * @param internalSubset Is this part of the internal or
     *  external subset?
     */
    public void addElementDecl(String name,
                               String contentSpec,
                               boolean internalSubset) {
        checkIfAlreadyCreated();
    }

    /**
     * Add an attribute declaration.
     * 
     * @param elementName The element name.
     * @param attrName The attribute name.
     * @param attrType The attribute type specification:
     *  CDATA, ID, IDREF, IDREFS, NMTOKEN, NMTOKENS,
     *  ENTITY, ENTITIES, NOTATION or ENUMERATION.
     * @param attrEnum - Enumeration for NOTATION and ENUMERATION.
     * @param defaultDecl The default value declaration,
     *  REQUIRED, IMPLIED, FIXED (or DEFAULT).
     * @param internalSubset Is this part of the internal or
     *  external subset?
     */
    public void addAttributeDecl(String elementName,
                                 String attrName,
                                 String attrType,
                                 String attrEnum,
                                 String defaultDecl,
                                 boolean internalSubset) {
        checkIfAlreadyCreated();
    }

    /**
     * Add an internal entity.
     *
     * @param name The entity name.
     * @param entityValue The value of the entity.
     * @param paramEntity Is this a parameter or general entity?
     * @param internalSubset Is this part of the internal or
     *  external subset?
     */
    public void addInternalEntityDecl(String name,
                                      String entityValue,
                                      boolean paramEntity,
                                      boolean internalSubset) {
        checkIfAlreadyCreated();
    }

    /**
     * Add an external entity.
     *
     * @param name The entity name.
     * @param systemId Document type system id.
     * @param publicId Document type public id, or null if not specified.
     * @param paramEntity Is this a parameter or general entity?
     * @param internalSubset Is this part of the internal or
     *  external subset?
     */
    public void addExternalEntityDecl(String name,
                                      String systemId,
                                      String publicId,
                                      boolean paramEntity,
                                      boolean internalSubset) {
        checkIfAlreadyCreated();
    }
    
    /**
     * Add an unparsed entity.
     *
     * @param name The entity name.
     * @param notationName The notation the entity references.
     * @param internalSubset Is this part of the internal or
     *  external subset?
     */
    public void addUnparsedEntityDecl(String name,
                                      String notationName,
                                      boolean internalSubset) {
        checkIfAlreadyCreated();
    }

    /**
     * Add a notation.
     *
     * @param name The notation name.
     * @param systemId Document type system id.
     * @param publicId Document type public id, or null if not specified.
     * @param internalSubset Is this part of the internal or
     *  external subset?
     */
    public void addNotationDecl(String name,
                                String systemId,
                                String publicId,
                                boolean internalSubset) {
        checkIfAlreadyCreated();
    }

    /**
     * Add internal subset as a single string.
     */
    public void setInternalSubset(String subsetStr) {
        checkIfAlreadyCreated();
        internalSubsetStr = subsetStr;
    }

    /**
     * Get the internal subset as a single string.
     */
    public String getInternalSubset() {
        return internalSubsetStr;
    }
}
