/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DOMBuilderGenerator.java,v 1.1.1.1 2003/03/10 16:36:22 taweili Exp $
 */

package org.enhydra.xml.xmlc.dom.generic;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.codegen.*;
import org.enhydra.xml.xmlc.dom.*;
import org.enhydra.xml.xmlc.compiler.ElementInfo;
import org.enhydra.xml.xmlc.compiler.ElementTable;
import org.enhydra.xml.dom.DOMOps;
import java.util.ArrayList;
import java.util.TreeMap;
import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;
import java.lang.Comparable;
import org.w3c.dom.*;

// FIXME: this code is still too ugly

/**
 * Class to generate code to build the document tree using only the W3C DOM
 * interface.  The document building code is generated in several functions to
 * avoid the JVM limit on maximum method size.  Since it is not possible to
 * determine the amount of bytecode generate, a metric is defined called
 * <em>create-cost</em>.  A unit of create-cost is approximately the cost
 * to create a node and append it to it's parent.  The limit on create-cost
 * is determined experimentally.
 *
 */
public class DOMBuilderGenerator {
    /** Add comments to generated code with node cost info */
    private static final boolean COST_DEBUG = false;

    /** Name of argument to method containing parent node. */
    protected static final String PARENT_NODE_ARG = "parentNode";

    /* 
     * Name of argument to method containing document.  Also used for local
     * variable name.
     */
    protected static final String DOCUMENT_ARG = "document";

    /* Name of variable containing document type */
    protected static final String DOCUMENT_TYPE_VAR = "docType";

    /** Maximum creation-cost that a single build method can handle. */
    private int fMaxCreateCostPerBuildMethod;

    /**
     * Type of document argument of build methods.  Allows derived
     * DOMs to use a more specific argument type to avoid casting.
     */
    private String fDocumentArgClassName;

    /** Name of method to create. */
    private String fMethodName;

    /** Source document. */
    private Document fDocument;

    /** Object used to create DOM nodes. */
    private NodeCreateGenerator fNodeCreator;

    /** Table of element information. */
    private ElementTable fElementTable;

    /**
     * Object that determines where new methods will be created to
     * prevent method overflow.
     */
    private BuildMethodMappings fBuildMethodMappings;

    /** Class being created. */
    private JavaClass fDocClass;

    /**
     * Generator of access methods, needed to generate access method
     * initialization
     */
    private AccessorGenerator fAccessorGenerator;

    /**
     * Should the methods be static.
     */
    private boolean fStaticMethods;

    /**
     * Class that represents a single subdocument building method.
     * <P>
     * A build method build either a subtree or a set of nodes with the
     * same parent. This is required because the stack for building the tree
     * is contained in variables in the build methods.  The same-parent
     * is used when all of the immediate children of a node will not fit
     * under the limit, these objects are refered to as being chained to
     * the object that created the parent.
     * <P>
     * The method that is generated that is associated with the
     * Document has the signature in the form:
     * <PRE>
     *   Document buildIt();
     * </PRE>
     * Others have
     * <PRE>
     *   void buildItSub1(Document doc, Node parentNode);
     * </PRE>
     * With <CODE>Document</CODE> being the specific type for the document.
     */
    private class BuildMethodGenerator {
        /** Table of tmp node variable names. */
        private VarNames fNodeVarNames = new VarNames("Node", "$node");

        /* Table of tmp Element node variable names. */
        private VarNames fElementVarNames = new VarNames("Element", "$elem");

        /* Table of tmp attribute variable names. */
        private VarNames fAttrVarNames = new VarNames("Attr", "$attr");

        /* Has document element variable been referenced? */
        private boolean fReferencedDocElement;

        /* The java code for the build metho */
        private JavaMethod fMethod;

        /** Body of current method being generated. */
        private JavaCode fBody;

        /** Next number to for generating submethod names. */
        private int fNextSubMethodId;

        /** If this is a chained entry, this is the chain level. */
        private int fChainLevel;

        /** The cost of this method. */
        private int fMethodCost;

        /**
         * Constructor. Create a method to generate a subtree and its
         * children.
         *
         * @param methodName The name for the build method.  A unique
         *  suffix is append to the name for each sub-method generated.
         */
        public BuildMethodGenerator(String methodName,
                                    Node root) throws XMLCException {
            // Initialize java method, with differente signiture for doc
            String type;
            JavaParameter[] params;
            if (root.getNodeType() ==  Node.DOCUMENT_NODE) {
                type = root.getClass().getName();
                params = new JavaParameter[0];
            } else {
                type = "void";
                JavaParameter docArg
                    = new JavaParameter(DOCUMENT_ARG,
                                        fDocumentArgClassName,
                                        (String)null);
                JavaParameter parentArg
                    = new JavaParameter(PARENT_NODE_ARG,
                                        Node.class.getName(),
                                        (String)null);
                params = new JavaParameter[]{docArg, parentArg};
            }
            fMethod = new JavaMethod(methodName, type,
                                     (JavaModifiers.PRIVATE | (fStaticMethods ? JavaModifiers.STATIC : 0)),
                                     params, new String[] {"Create a subtree of the document."});
            fDocClass.addMethod(fMethod);
            fBody = fMethod.getCode();
        }

        /**
         * Create a call to the method represented by this object.  Not used
         * for top level.
         */
        public void createMethodCall(String docVar,
                                     String parentVar,
                                     JavaCode body) {
            body.addln(fMethod.getName() + "(" + docVar + ", "
                       + parentVar + ");");
        }

        /**
         * Generate a method to create DOM.  This is the document-level 
         * entry point,
         * @param doc The document
         * @param body Append a call to this build method here, if not null.
         */
        public void createMethod(Document document) throws XMLCException {
            if (COST_DEBUG) {
                fBody.addln("// " +fBuildMethodMappings.toString(document));
            }

            // This creates the document and document type, also a local
            // variables holding the document and document type.
            fNodeCreator.genDocumentCreate(document, DOCUMENT_ARG,
                                           DOCUMENT_TYPE_VAR, fBody);
            fMethodCost += fBuildMethodMappings.getNodeTypeCost(document);
            
            // Build whole main tree
            processChildren(DOCUMENT_ARG, document, 1);

            // Build document type
            DocumentType docType = document.getDoctype();
            if (docType != null) {
                processDocumentType(DOCUMENT_TYPE_VAR, docType, 1);
            }

            fBody.addln("return " + DOCUMENT_ARG + ";");
            createVariables();

            if (fMethodCost > fMaxCreateCostPerBuildMethod) {
                throw new XMLCError("BUG: maximum number of nodes per method exceeded");
            }
        }
        
        /**
         * Generate a method to create DOM nodes.  This is the non-document
         * entry point, it may recursively create more methods.
         * @param root The root of the subtree. 
         * @param docVar The variable containing the document, if not null.
         */
        public void createMethod(Node parent,
                                 Node root) throws XMLCException {
            processNode(PARENT_NODE_ARG, parent, root, 0);
            createVariables();

            if (fMethodCost > fMaxCreateCostPerBuildMethod) {
                throw new XMLCError("BUG: maximum number of nodes per method exceeded");
            }
        }
        
        /**
         * Create children of a node using chaining.  This is the entry when
         * creating more chained nodes.  
         * @param children the enumeration should point to the next child
         *  to add.
         */
        private void createChainedChildrenMethod(int chainLevel,
                                                 String parentVar,
                                                 Node parent,
                                                 Node nextChild,
                                                 String docVar,
                                                 JavaCode body) throws XMLCException {
            fChainLevel = chainLevel;
            body.addln(fMethod.getName() + "(" + docVar + ", " + parentVar + ");");
            processChildrenChaining(PARENT_NODE_ARG, parent, nextChild, 0);
            createVariables();
        }

        /**
         * Get temporary variable name for a node. Special handling for
         * Elements; allow for attribute setting without requiring (slower)
         * type casting in generate code.
         */
        private String getNodeVarName(Node node,
                                      int level) {
            if (node instanceof Element) {
                return fElementVarNames.getVarName(level);
            } else {
                return fNodeVarNames.getVarName(level);
            }
        }

        /**
         * Get the name of the next subtree method to build.
         */
        private String getNextSubTreeMethodName() {
            return fMethod.getName() + "_" + fNextSubMethodId++;
        }

        /**
         * Determine how much space is left in the method, in create-cost units.
         */
        private int spaceLeft() {
            int left = (fMaxCreateCostPerBuildMethod - fMethodCost);
            return (left < 0) ? 0 : left;
        }

        /**
         * Create a new build method for a subtree.
         */
        private void createBuildMethod(String parentVar,
                                       Node parent,
                                       Node node) throws XMLCException {
            BuildMethodGenerator buildMethod
                = new BuildMethodGenerator(getNextSubTreeMethodName(), node);
            buildMethod.createMethodCall(DOCUMENT_ARG, parentVar, fBody);
            buildMethod.createMethod(parent, node);
            fMethodCost += BuildMethodMappings.BUILD_METHOD_CALL_CREATE_COST;
        }
        
        /**
         * Generate code to to initialize an attribute.
         */
        private void processAttr(String elementVar,
                                 Element element,
                                 Attr attr,
                                 int level) throws XMLCException {
            String attrVar = fAttrVarNames.getVarName(level);
            if (attr.getSpecified()) {
                // Create attribute node
                fNodeCreator.genNodeCreate(DOCUMENT_ARG, attrVar, attr, fBody);
                fNodeCreator.genAddAttribute(elementVar, attrVar, fBody);
                fMethodCost += fBuildMethodMappings.getNodeTypeCost(attr);

                // Create attribute children.
                processChildren(attrVar, attr, level+1);
            }
        }

        /**
         * Generate code to initialize an element's attributes.
         */
        private void processAttrs(String elementVar,
                                  Element element,
                                  int level) throws XMLCException {
            NamedNodeMap attrList = element.getAttributes();
            if (attrList != null) {
                for (int idx = 0; idx < attrList.getLength(); idx++) {
                    processAttr(elementVar, element,
                                (Attr)attrList.item(idx), level);
                }
            }
        }

        /**
         * Generate code to initialize element-specific constructs and create
         * its attributes.
         */
        private void processElement(String elementVar,
                                    Element element,
                                    int level) throws XMLCException {
            processAttrs(elementVar, element, level);
            ElementInfo elementInfo = fElementTable.getElementInfo(element);
            if (elementInfo.getNumAccessMethods() > 0) {
                fAccessorGenerator.createAccessMethodInit(elementInfo, elementVar,
                                                          fBody);
            }
        }

        /**
         * Generate code to initialize an entity in a DocumentType.
         */
        private void processEntity(String docTypeVar,
                                   Entity entity,
                                   int level) throws XMLCException {
            String entityVar = getNodeVarName(entity, level);
            fNodeCreator.genNodeCreate(DOCUMENT_ARG, entityVar, entity, fBody);
            fNodeCreator.genAddEntity(docTypeVar, entityVar, entity, fBody);
            fMethodCost += fBuildMethodMappings.getNodeTypeCost(entity);
            processChildren(entityVar, entity, level+1);
        }

        /**
         * Generate code to initialize an notation in a DocumentType.
         */
        private void processNotation(String docTypeVar,
                                     Notation notation,
                                     int level) throws XMLCException {
            String notationVar = getNodeVarName(notation, level);
            fNodeCreator.genNodeCreate(DOCUMENT_ARG, notationVar, notation, fBody);
            fNodeCreator.genAddNotation(docTypeVar, notationVar, notation, fBody);
            fMethodCost += fBuildMethodMappings.getNodeTypeCost(notation);
        }

        /**
         * Generate code to initialize DocumentType contained nodes.
         */
        private void processDocumentType(String docTypeVar,
                                         DocumentType docType,
                                         int level) throws XMLCException {
            NamedNodeMap entities = docType.getEntities();
            if (entities != null) {
                for (int idx = 0; idx < entities.getLength(); idx++) {
                    processEntity(docTypeVar, (Entity)entities.item(idx),
                                  level);
                }
            }
            NamedNodeMap notations = docType.getNotations();
            if (notations != null) {
                for (int idx = 0; idx < notations.getLength(); idx++) {
                    processNotation(docTypeVar, (Notation)notations.item(idx),
                                    level);
                }
            }
        }

        /**
         * Generate code to create a node of most types.  This doesn't
         * handle Document or attributes.
         * @return The generated variable name that will contain the node.
         */
        private String createNode(String parentVar,
                                  Node parent,
                                  Node node,
                                  int level) throws XMLCException {
            String nodeVar = getNodeVarName(node, level);
            fNodeCreator.genNodeCreate(DOCUMENT_ARG, nodeVar, node, fBody);
            fNodeCreator.genAppendChild(parentVar, nodeVar, node, fBody);
            fMethodCost += fBuildMethodMappings.getNodeTypeCost(node);
            fBody.addln();
            return nodeVar;
        }

        /**
         * Process a node, generating code to recreate it.
         */
        private void processNode(String parentVar,
                                 Node parent,
                                 Node node,
                                 int level) throws XMLCException {
            if (node instanceof DocumentType) {
                return;  // already handled
            }
            if ((level != 0) && fBuildMethodMappings.isMethodRoot(node)) {
                // Make this element root of a method
                createBuildMethod(parentVar, parent, node);
            } else {
                if (COST_DEBUG) {
                    fBody.addln("// " + fBuildMethodMappings.toString(node));
                }
                String nodeVar = createNode(parentVar, parent, node, level);
                if (node instanceof Element) {
                    processElement(nodeVar, (Element)node, level);
                }
                processChildren(nodeVar, node, level+1);
            }
        }

        /**
         * Process nodes contained in parent node while chaining build methods
         * together.  This fills in the current build method and then chains
         * to another.
         */
        private void processChildrenChaining(String parentVar,
                                             Node parent,
                                             Node nextChild,
                                             int level) throws XMLCException {
            // Fit as many as we an into this method.  At least fit one, to
            // handle the case of the calculated cost exceeding the max cost
            // of a method.
            boolean isFirst = true;
            while ((nextChild != null)
                   && ((fBuildMethodMappings.getCreateCost(nextChild) < spaceLeft())
                       || isFirst)) {
                processNode(parentVar, parent, nextChild, level);
                nextChild = nextChild.getNextSibling();
                isFirst = false;
            }
            
            // If there are any left, create another method.
            if (nextChild != null) {
                BuildMethodGenerator nextBuildMethod
                    = new BuildMethodGenerator(fMethod.getName() + "c",
                                               nextChild);
                nextBuildMethod.createChainedChildrenMethod(fChainLevel+1,
                                                            parentVar, parent, 
                                                            nextChild,
                                                            DOCUMENT_ARG,
                                                            fBody);
                fMethodCost += BuildMethodMappings.BUILD_METHOD_CALL_CREATE_COST;
            }
        }

        /**
         * Create children of the node.
         */
        private void processChildren(String parentVar,
                                     Node parent,
                                     int level) throws XMLCException {
            if (fBuildMethodMappings.useChainedChildrenMethods(parent)) {
                processChildrenChaining(parentVar, parent,
                                        parent.getFirstChild(), level);
            } else {
                for (Node child = parent.getFirstChild();
                     child != null; child = child.getNextSibling()) {
                    processNode(parentVar, parent, child, level);
                }
            }
        }

        /**
         * Fill in the variables for this method.
         */
        private void createVariables() {
            fNodeVarNames.insertVarDefs(fBody);
            fElementVarNames.insertVarDefs(fBody);
            fAttrVarNames.insertVarDefs(fBody);
            fBody.addVars(""); // blank line
        }
    }

    /**
     * Constructor.  Build the methods.
     */
    public DOMBuilderGenerator(String methodName,
                               Document document,
                               String documentArgClassName,
                               NodeCreateGenerator nodeCreator,
                               AccessorGenerator accessorGenerator,
                               ElementTable elementTable,
                               JavaClass docClass,
                               int maxCreateCostPerBuildMethod,
                               boolean staticMethods) throws XMLCException {
        fMethodName = methodName;
        fDocument = document;
        fDocumentArgClassName = documentArgClassName;
        fNodeCreator = nodeCreator;
        fAccessorGenerator = accessorGenerator;
        fElementTable = elementTable;
        fDocClass = docClass;
        fMaxCreateCostPerBuildMethod = maxCreateCostPerBuildMethod;
        fStaticMethods = staticMethods;

        fBuildMethodMappings = new BuildMethodMappings(fMaxCreateCostPerBuildMethod,
                                                       document);

        BuildMethodGenerator buildMethod
            = new BuildMethodGenerator(fMethodName, document);
        buildMethod.createMethod(document);
    }

    /**
     * Create a call to the top level method that was generated.
     */
    public void createMethodCall(JavaCode body) {
        body.addln(fMethodName + "();");
    }
}
