/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: GenericAccessorGenerator.java,v 1.2 2003/06/29 05:58:48 jkjome Exp $
 */

package org.enhydra.xml.xmlc.dom.generic;

import org.enhydra.xml.xmlc.dom.AccessorGenerator;
import org.enhydra.xml.xmlc.compiler.ElementInfo;
import org.enhydra.xml.xmlc.compiler.ClassGenerator;
import org.enhydra.xml.xmlc.codegen.JavaClass;
import org.enhydra.xml.xmlc.codegen.JavaMethod;
import org.enhydra.xml.xmlc.codegen.JavaField;
import org.enhydra.xml.xmlc.codegen.JavaCode;
import org.enhydra.xml.xmlc.codegen.JavaParameter;
import org.enhydra.xml.xmlc.codegen.JavaModifiers;

/**
 * Generate access methods for a DOM only using the W3C DOM.
 * Many methods in the class are protected so it can be used
 * as a base for other AccessorGenerator classes.
 */
public class GenericAccessorGenerator implements AccessorGenerator {
    /**
     * Parameter for setTextXXX methods.
     * A static since in can be shared.
     */
    private static final JavaParameter[] setTextParam = {
        new JavaParameter("text", "String",
                          "New value for text child.")
    };

    /**
     * Get the name of the field that will store an element.
     */
    protected String getElementField(ElementInfo elementInfo) {
        return "$element_" + elementInfo.getJavaId();
    }

    /**
     * Create code to setup an access method for an element when building the
     * DOM.
     * @see AccessorGenerator#createAccessMethodInit
     */
    public void createAccessMethodInit(ElementInfo elementInfo,
                                       String elementVarName,
                                       JavaCode body) {
        body.add(getElementField(elementInfo) 
                 + " = (" + elementInfo.getReferenceName() + ")"
                 + elementVarName + ";");
    }

    /**
     * Create code cached information so that element access methods
     * return null.
     * @see AccessorGenerator#createNullElementAccess
     */
    public void createNullElementAccess(ElementInfo elementInfo,
                                        JavaCode body) {
        body.add(getElementField(elementInfo) + " = null;");
    }

    /**
     * Create code to reset the element access information from a node.
     * @see AccessorGenerator#createResetElementAccess
     */
    public void createResetElementAccess(ElementInfo elementInfo,
                                         String nodeVarName,
                                         JavaCode body) {
        body.addln(getElementField(elementInfo) + " = ("
                 + elementInfo.getReferenceName() + ")" + nodeVarName + ";");
    }

    /**
     * Set up a accessor (getElementXXX or getTagXXX) method object for an
     * element.  Creates the method, documentation, and delegate check code.
     */
    protected JavaMethod setupGetElementMethod(ElementInfo elementInfo,
                                               ElementInfo.AccessorInfo accessor,
                                               boolean createDelegation,
                                               JavaClass docClass) {
        String[] doc = {
            "Get the element with id <CODE>" + elementInfo.getXmlId() + "</CODE>.",
            "@see " + accessor.fReturnType,
        };
        JavaMethod method
            = new JavaMethod(accessor.fName,
                             accessor.fReturnType,
                             JavaModifiers.PUBLIC,
                             null, doc);
        docClass.addMethod(method);

        // Setup delegate check and call, if requested.
        JavaCode body = method.getCode();
        if (createDelegation) {
            body.addln(new String[] {
                "if (" + ClassGenerator.DELEGATE_FIELD_NAME + " != null) {",
                "    return " + ClassGenerator.DELEGATE_FIELD_NAME + "." + method.getName() + "();",
                "}"
            });
        }
        return method;
    }

    /**
     * Create an element accessor method (getElementXXX() or getTextXXX()).
     * @see AccessorGenerator#createAccessorMethod
     */
    public JavaMethod createAccessorMethod(ElementInfo elementInfo,
                                           ElementInfo.AccessorInfo accessor,
                                           boolean createDelegation,
                                           JavaClass docClass) {
        JavaMethod method = setupGetElementMethod(elementInfo, accessor,
                                                  createDelegation,
                                                  docClass);

        // Create return of value
        JavaCode body = method.getCode();
        body.addln("return " + getElementField(elementInfo) +  ";");

        // Create field to store element.
        JavaField field = new JavaField(getElementField(elementInfo),
                                        elementInfo.getReferenceName(),
                                        JavaModifiers.PRIVATE,
                                        (String)null, null);
        docClass.addField(field);
        return method;
    }

    /**
     * Setup the setText method for an element.
     * Creates the method, documentation, and delegate check code.
     */
    protected JavaMethod setupSetTextMethod(ElementInfo elementInfo,
                                            boolean createDelegation,
                                            JavaClass docClass) {
        String[] doc = {
            "Get the value of text child of element <CODE>" + elementInfo.getXmlId() + "</CODE>.",
            "@see org.w3c.dom.Text"
        };

        JavaMethod method
            = new JavaMethod("setText" + elementInfo.getJavaId(),
                             "void",
                             JavaModifiers.PUBLIC,
                             setTextParam, doc);
        docClass.addMethod(method);

        // Setup delegate check and call, if requested.
        JavaCode body = method.getCode();
        if (createDelegation) {
            body.addln(new String[] {
                "if (" + ClassGenerator.DELEGATE_FIELD_NAME + " != null) {",
                "    " + ClassGenerator.DELEGATE_FIELD_NAME + "." + method.getName() + "(text);",
                "    return;",
                "}"
            });
        }
        return method;
    }

    /**
     * Create the setText method for an element.
     * @see AccessorGenerator#createSetTextMethod
     */
    public JavaMethod createSetTextMethod(ElementInfo elementInfo,
                                          boolean createDelegation,
                                          JavaClass docClass) {
        JavaMethod method = setupSetTextMethod(elementInfo,
                                               createDelegation,
                                               docClass);

        // Create set of value
        JavaCode body = method.getCode();
        body.addln("doSetText(" + getElementField(elementInfo) + ", text);");
        return method;
    }
}
