/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: TidyHTMLParser.java,v 1.1.1.1 2003/03/10 16:36:22 taweili Exp $
 */

package org.enhydra.xml.xmlc.html.parsers.tidy;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.html.*;
import org.enhydra.xml.xmlc.html.parsers.HTMLParserBase;
import org.enhydra.xml.xmlc.html.parsers.HTMLDocBuilder;
import org.enhydra.xml.xmlc.compiler.*;
import org.enhydra.xml.xmlc.parsers.ParseTracer;
import org.enhydra.xml.xmlc.parsers.XMLCParser;
import org.enhydra.xml.xmlc.metadata.MetaData;
import org.enhydra.xml.xmlc.metadata.HTMLSection;
import org.enhydra.xml.xmlc.metadata.HTMLTagDef;
import org.enhydra.xml.xmlc.metadata.HTMLAttrDef;
import org.enhydra.xml.xmlc.metadata.ParserType;
import org.enhydra.xml.xmlc.dom.*;
import org.enhydra.xml.xmlc.misc.LineNumberMap;
import org.enhydra.xml.xmlc.misc.SSIReader;
import org.enhydra.xml.io.ErrorReporter;
import org.enhydra.xml.io.InputSourceOps;
import org.w3c.dom.*;
import org.w3c.dom.html.*;
import java.net.URL;
import java.io.*;
import java.util.StringTokenizer;
import java.util.Enumeration;
import org.w3c.tidy.*;
import org.xml.sax.InputSource;

/**
 * XMLCParser object for HTML and HTML framesets that uses the Java version of
 * the W3C HTML tidy program.  It uses Tidy to convert HTML to XHTML and then
 * parses it with an XML parser.
 */
public class TidyHTMLParser extends HTMLParserBase implements XMLCParser {
    /**
     * Verbose tracing.
     */
    private ParseTracer fTracer;

    /**
     * The document builder.
     */
    private HTMLDocBuilder fDocBuilder;

    /**
     * Tidy parser.
     */
    private Tidy fTidy = new Tidy();

    /**
     * Constructor.
     */
    public TidyHTMLParser() throws XMLCException {
        fTidy.setTidyMark(false); // Don't add Tidy META tag.
        fTidy.setQuiet(true);  // Doc print guess HTML version
    }

    /**
     * Get the type of a Tidy node as a string.
     */
    private String nodeTypeStr(short typeId) {
        switch (typeId) {
        case org.w3c.tidy.Node.RootNode: return "RootNode";
        case org.w3c.tidy.Node.DocTypeTag: return "DocTypeTag";
        case org.w3c.tidy.Node.CommentTag: return "CommentTag";
        case org.w3c.tidy.Node.ProcInsTag: return "ProcInsTag";
        case org.w3c.tidy.Node.TextNode: return "TextNode";
        case org.w3c.tidy.Node.StartTag: return "StartTag";
        case org.w3c.tidy.Node.EndTag: return "EndTag";
        case org.w3c.tidy.Node.StartEndTag: return "StartEndTag";
        case org.w3c.tidy.Node.AspTag: return "AspTag";
        default: return "**Unknown Node**";
        }
    }

    /**
     * Print information about a Tidy node.
     */
    private void printNodeInfo(org.w3c.tidy.Node tNode) {
        StringBuffer buf = new StringBuffer();
        buf.append(nodeTypeStr(tNode.getType()) + ":");
        if (tNode.getElement() != null) {
            buf.append(" " + tNode.getElement());
        }
        // Print attributes
        AttVal attVal = tNode.getAttributes();
        while (attVal != null) {
            buf.append("  " + attVal.attribute + "=\"" + attVal.value + "\"");
            attVal = attVal.next;
        }

        // Print associated text
        if (tNode.getType() != org.w3c.tidy.Node.StartTag) {
            String value = tNode.getNodeValue();
            if (value != null) {
                buf.append(" '" + value + "'");
            }
        }
        fTracer.trace(buf.toString());
    }

    /**
     * Recursively print information about a Tidy node.
     */
    private void printNode(org.w3c.tidy.Node tNode) {
        printNodeInfo(tNode);
        org.w3c.tidy.Node next = tNode.getContent();
        fTracer.enter();
        while (next != null) {
            printNode(next);
            next = next.getNext();
        }
        fTracer.leave();
    }

    /**
     * Create an Element node from a Tidy node.
     */
    private void createElement(org.w3c.tidy.Node tNode) {
        if (tNode.getElement() == null) {
            return;
        }
        fDocBuilder.startElement(tNode.getElement());
        
        AttVal attVal = tNode.getAttributes();
        while (attVal != null) {
            String value = (attVal.value == null) ? "" : attVal.value;
            fDocBuilder.addAttribute(attVal.attribute, value);
            attVal = attVal.next;
        }
    }

    /**
     * Create a DOM node from a Tidy node.
     * 
     * @param tNode Tidy node.
     */
    private void makeDomNode(org.w3c.tidy.Node tNode) {
        switch (tNode.getType()) {
        case org.w3c.tidy.Node.CommentTag:
            if (tNode.getNodeValue() != null) {
                fDocBuilder.addComment(tNode.getNodeValue());
            }
            break;
        case org.w3c.tidy.Node.TextNode:
            if (tNode.getNodeValue() != null) {
                fDocBuilder.addTextNode(tNode.getNodeValue());
            }
            break;
        case org.w3c.tidy.Node.StartTag:
        case org.w3c.tidy.Node.StartEndTag:;
            createElement(tNode);
            break;
        case org.w3c.tidy.Node.EndTag:
            new XMLCError("Internal error: Unexpected Tidy EndTag node");
            break;
        case org.w3c.tidy.Node.RootNode:
        case org.w3c.tidy.Node.DocTypeTag:
        case org.w3c.tidy.Node.ProcInsTag:
        case org.w3c.tidy.Node.AspTag:
            break;
        default: 
            throw new XMLCError("Internal error: Unknown node");
        }
    }

    /**
     * Recursively convert a node of the Tidy tree to a DOM tree.
     */
    private void buildNode(org.w3c.tidy.Node tNode) {

        makeDomNode(tNode);

        // Recurse to process children.
        org.w3c.tidy.Node tNext = tNode.getContent();
        while (tNext != null) {
            buildNode(tNext);
            tNext = tNext.getNext();
        }
        if (tNode.getType() == org.w3c.tidy.Node.StartTag) {
            fDocBuilder.finishElement();
        }
    }

    /**
     * Check for a Tidy-supported encoding.  If one is found,
     * set it as the encoding.
     * @param htmlEncoding Encoding to check, null is assumed to
     *  be ASCII.
     * @return true if encoding is Tidy-supported, false if its
     *  not.
     */
    private boolean checkForTidyEncoding(String htmlEncoding) {
        if ((htmlEncoding == null)
            || (htmlEncoding.equalsIgnoreCase("US-ASCII"))
            || (htmlEncoding.equalsIgnoreCase("ASCII"))) {
            fTidy.setCharEncoding(Configuration.ASCII);
            return true;
        } else if (htmlEncoding.equalsIgnoreCase("ISO-8859-1")) {
            fTidy.setCharEncoding(Configuration.LATIN1);
            return true;
        } else if (htmlEncoding.equalsIgnoreCase("UTF-8")) {
            fTidy.setCharEncoding(Configuration.UTF8);
            return true;
        } else if (htmlEncoding.equalsIgnoreCase("ISO-2022-JP")) {
            fTidy.setCharEncoding(Configuration.ISO2022);
            return true;
        } else {
            return false;
        }
    }

    /**
     * Convert the input to a UTF-8 byte array and make it into
     * an input stream.
     */
    private InputStream makeUTF8InputStream(InputSource input)
        throws IOException {
        
        // Read into a UTF byte array.
        Reader reader = InputSourceOps.open(input);
        try {
            ByteArrayOutputStream utf8Bytes = new ByteArrayOutputStream();
            OutputStreamWriter writer = new OutputStreamWriter(utf8Bytes,
                                                               "UTF8");
            char buffer[] = new char[4096];
            int readSize;
            while ((readSize = reader.read(buffer)) >= 0) {
                writer.write(buffer, 0, readSize);
            }
            writer.flush();

            // Now we can read it with Tidy
            fTidy.setCharEncoding(Configuration.UTF8);
            return new ByteArrayInputStream(utf8Bytes.toByteArray());
            
        } finally {
            InputSourceOps.closeIfOpened(input, reader);
        }
    }

    /**
     * Get the input stream.  If the specified encoding is not one supported
     * by Jtidy or a character stream is open, the file is read into a byte
     * array of UTF characters and an input stream wrapped around that array.
     * FIXME: If JTidy work on character streams, this mess could all go
     * away.
     */
    private InputStream getInputStream(InputSource input) throws IOException {
        String htmlEncoding = input.getEncoding();
        if (!checkForTidyEncoding(htmlEncoding)) {
            // Encoding not supported by tidy.
            return makeUTF8InputStream(input);
        }
        if (input.getByteStream() != null) {
            // Byte stream already open, use it.
            return input.getByteStream();
        }
        if (input.getCharacterStream() != null) {
            // Need to convert back to a byte stream.
            return makeUTF8InputStream(input);
        }
        // Open it ourselves.
        return InputSourceOps.openSystemId(input.getSystemId());
    }

    /**
     * Specify the list of proprietary tags and attributes.
     */
    private void setProprietaryTags(HTMLSection htmlSection) throws XMLCException {
        //FIXME: Ouch, this is global to tidy; problem if invoking multiple times.

        HTMLTagDef[] tagDefs = htmlSection.getHTMLTagDefs();
        TagTable tagTable = fTidy.getConfiguration().getTagTable();
        for (int idx = 0; idx < tagDefs.length; idx++) {
            addTag(tagTable, tagDefs[idx]);
        }
    
        HTMLAttrDef[] attrDefs = htmlSection.getHTMLAttrDefs();
        AttributeTable attributeTable = AttributeTable.getDefaultAttributeTable();
        for (int idx = 0; idx < attrDefs.length; idx++) {
            attributeTable.install(new Attribute(attrDefs[idx].getName().toLowerCase(),
                                                 Dict.VERS_PROPRIETARY,
                                                 null));
        }
    }

    /**
     * @see XMLCParser#parse
     */
    public XMLCDocument parse(InputSource input,
                              LineNumberMap lineNumberMap,
                              XMLCDomFactory domFactory,
                              MetaData metaData,
                              ErrorReporter errorReporter,
                              ParseTracer tracer)
        throws IOException, XMLCException {

        validateConf(ParserType.TIDY, metaData);

        fTracer = tracer;
        fDocBuilder = new HTMLDocBuilder(domFactory, input);

        setProprietaryTags(metaData.getHTMLSection());
        fTidy.setInputStreamName(input.getSystemId());
        fTidy.setErrout(new TidyErrorHandler(errorReporter,
                                             input.getSystemId(),
                                             lineNumberMap));

        InputStream srcFileStream = getInputStream(input);
        org.w3c.tidy.Node tRoot;
        try {
            tRoot = fTidy.parse(srcFileStream, null);
        } finally {
            // Close if was not open in InputSource
            if (!InputSourceOps.isOpen(input)) {
                srcFileStream.close();
            }
        }
        //FIXME: need to get error count from parser..
        if (errorReporter.getErrorCnt() != 0) {
            handleParseErrors(errorReporter);
        }
        
        if ((fTracer != null) && fTracer.enabled()) {
            printNode(tRoot);
        }

        // Build DOM
        buildNode(tRoot);
        addPCDataContentElements(fDocBuilder.getXMLCDocument());
        return fDocBuilder.getXMLCDocument();
    }

    /**
     * Add a proprietary tag to the set of allowed tags.
     */
    private void addTag(TagTable tagTable,
                        HTMLTagDef tagDef) throws XMLCException {
        Parser tagParser = null;
        int model = 0;

        // Get basic content model, alow multiple to be combine,
        // although this only makes sense with empty and the others.
        if (tagDef.getEmpty()) {
            model |= Dict.CM_EMPTY|Dict.CM_OPT;
            tagParser = ParserImpl.getParseInline();
        }
        if (tagDef.getInline()) {
            model |= Dict.CM_INLINE;
            tagParser =  ParserImpl.getParseInline();
        }
        if (tagDef.getBlock()) {
            model |= Dict.CM_BLOCK;
            tagParser = ParserImpl.getParseBlock();
        }
        if (model == 0) {
            throw new XMLCException("must specify at least one on TAG_CM_EMPTY, TAG_CM_INLINE, or TAG_CM_BLOCK");
        }
        
        // Is close optional?
        if (tagDef.getOptclose()) {
            model |= Dict.CM_OPT;
        }

        // Let them occur anywhere and contain anything
        model |= Dict.CM_HEAD|Dict.CM_HTML|Dict.CM_MIXED;

        tagTable.install(new Dict(tagDef.getName().toLowerCase(),
                                  Dict.VERS_PROPRIETARY,
                                  model, tagParser, null));
    }
}
