/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DOMType.java,v 1.2 2003/08/02 05:02:48 jkjome Exp $
 */

package org.enhydra.xml.xmlc.metadata;

/**
 * Enumerated type for the predefined DOM factories.
 */
public class DOMType extends EnumeratedType {
    /**
     * Constant for LazyDOM.
     */
    public static final DOMType LAZYDOM
        = new DOMType("lazydom",
                      org.enhydra.xml.xmlc.dom.lazydom.LazyDomFactory.class,
                      org.enhydra.xml.xmlc.dom.lazydom.LazyHTMLDomFactory.class);

    /**
     * Constant for Xerces DOM.
     */
    public static final DOMType XERCES
        = new DOMType("xerces",
                      org.enhydra.xml.xmlc.dom.xerces.XercesDomFactory.class,
                      org.enhydra.xml.xmlc.dom.xerces.XercesHTMLDomFactory.class);


    /**
     * Constant for genric DOM.
     */
    // FIXME: Check if xerces DOM really is generic enough...
    public static final DOMType GENERIC
        = new DOMType("generic",
                      org.enhydra.xml.xmlc.dom.xerces.XercesDomFactory.class,
                      org.enhydra.xml.xmlc.dom.xerces.XercesHTMLDomFactory.class);

    /**
     * Constant for XHTML DOM.
     */
    public static DOMType XHTML;

    /**
     * Constant for WML DOM.
     */
    public static DOMType WML;
    
    /**
     * Constant for VoiceXML DOM.
     */
    public static DOMType VOICEXML;
	
    /**
     * Constant for cHTML DOM.
     */
    public static DOMType CHTML;

    // Initialize the optional DOM types
    static {
	try {
	    XHTML = new DOMType("xhtml",
				Class.forName("org.enhydra.xml.xhtml.XHTMLDomFactory"),
				Class.forName("org.enhydra.xml.xhtml.HTMLDomFactory"));
	} catch (ClassNotFoundException e) { XHTML = null; }

	try {
	    WML = new DOMType("wml",
			      Class.forName("org.enhydra.wireless.wml.WMLDomFactory"),
			      null);
	} catch (ClassNotFoundException e) { WML = null; }

	try {
	    VOICEXML = new DOMType("voicexml",
				   Class.forName("org.enhydra.wireless.voicexml.VoiceXMLDomFactory"),
				   null);

	} catch (ClassNotFoundException e) { VOICEXML = null; }

	try {
	    CHTML = new DOMType("chtml",
				null,
				Class.forName("org.enhydra.wireless.chtml.CHTMLDomFactory"));
	} catch (ClassNotFoundException e) { CHTML = null; }
    }

    /**
     * XML DOM factory class for this type.
     */
    private final Class fXMLDomFactoryClass;

    /**
     * HTML DOM factory class for this type.
     */
    private final Class fHTMLDomFactoryClass;

    /**
     * Internal constructor.  Only a fixed set of types are allowed.
     */
    private DOMType(String name,
                    Class xmlDomFactoryClass,
                    Class htmlDomFactoryClass) {
        super(name);
        fXMLDomFactoryClass = xmlDomFactoryClass;
        fHTMLDomFactoryClass = htmlDomFactoryClass;
    }

    /**
     * Get the XML DOM factory class for this type.
     */
    public Class getXMLDomFactoryClass() {
        return fXMLDomFactoryClass;
    }

    /**
     * Get the HTML DOM factory class for this type.
     */
    public Class getHTMLDomFactoryClass() {
        return fHTMLDomFactoryClass;
    }

    /**
     * Get the object for the specified type.  If desiredType is null,
     * return null.
     */
    public static DOMType getType(String desiredType) {
        if (desiredType == null) {
            return null;
        } else if (LAZYDOM.fName.equals(desiredType)) {
            return LAZYDOM;
        } else if (XERCES.fName.equals(desiredType)) {
            return XERCES;
        } else if (XHTML != null && XHTML.fName.equals(desiredType)) {
            return XHTML;
        } else if (CHTML != null && CHTML.fName.equals(desiredType)) {
            return CHTML;
        } else if (VOICEXML != null && VOICEXML.fName.equals(desiredType)) {
            return VOICEXML;
        } else if (WML != null && WML.fName.equals(desiredType)) {
            return WML;
        } else {
            throw new IllegalArgumentException(
			   "Invalid DOMType: \""
			   + desiredType
			   + "\", expected one of \""
			   + (XHTML!=null?XHTML.fName + "\", \"":"")
			   + (WML!=null?WML.fName + "\", \"":"")
			   + (VOICEXML!=null?VOICEXML.fName + "\", \"":"")
			   + (CHTML!=null?CHTML.fName + "\", \"":"")
			   + LAZYDOM.fName + "\" or \""
			   + XERCES.fName);
        }
    }
}
