/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: TidyTagInfo.java,v 1.1.1.1 2003/03/10 16:36:25 taweili Exp $
 */

package org.enhydra.xml.devtools.tidy;

import org.enhydra.xml.xmlc.*;
import org.w3c.tidy.*;
import java.util.Hashtable;
import java.util.Enumeration;
import java.io.*;


/**
 * Class and program to dump information about each tag from the tidy tag table.
 */
public class TidyTagInfo {
    /**
     * Content model table.
     */
    private static final ContentModelTable contentModelTable = new ContentModelTable();

    /**
     * Content model columns for the table.
     */
    private static int[] columnContentModels;

    /**
     * Static constructor.
     */
    static {
        columnContentModels = new int[contentModelTable.getSize()];
        Enumeration mods = contentModelTable.getNumbers();
        int idx = 0;

        while (mods.hasMoreElements()) {
            columnContentModels[idx++] = ((Integer)mods.nextElement()).intValue();
        }
    }


    /**
     * Print header for table.
     */
    private void printHeader(PrintWriter out,
                             int contentModel) {
        String contentModelStr = contentModelTable.getSymbol(contentModel);
        out.println("<HTML><HEAD><TITLE>"
                    + "Tags order by " + contentModelStr
                    + "</TITLE></HEAD><BODY BF='#FFFFFF'><H1>"
                    + "Tags order by " + contentModelStr
                    + " </H1>");
        out.println("<TABLE BORDER='1'>");
        out.println("<TR ALIGN='CENTER'>");
        out.print("<TH>Name");
        int maxLen = 6;
        for (int idx = 0; idx < columnContentModels.length; idx++) {
            String sym = contentModelTable.retrieveSymbol(columnContentModels[idx]);
            sym = sym.substring(0, (sym.length() > maxLen) ? maxLen : sym.length());
            out.print("<TH>");
            for (int cnt = 0; cnt < maxLen-sym.length(); cnt++) {
                out.print("&nbsp;");
                out.print("<BR>");
            }
            for (int ci = 0; ci < sym.length(); ci++) {
                out.print(sym.charAt(ci));
                out.print("<BR>");
            }
            out.println();
        }
        out.println("</TR>");
    }

    /**
     * Print trailer for table and page.
     */
    private void printTrailer(PrintWriter out) {
        out.println("</TABLE>");
        out.println("</BODY></HTML>");
    }

    /**
     * Print info about a tag.
     */
    private void printTag(PrintWriter out,
                          Dict tag) {
        out.print("<TR ALIGN='CENTER'>");
        out.print("<TD>");
        out.print(tag.name);
        for (int idx = 0; idx < columnContentModels.length; idx++) {
            out.print("<TD>");
            if ((tag.model & columnContentModels[idx]) != 0) {
                out.print("<FONT COLOR='#FF0000'>*</FONT>");
            } else {
                out.print("&nbsp;");
            }
        }
        out.println("</TR>");
    }

    /**
     * Generate a tag entries including or excluding the specified content model.
     */
    private void printTagEntries(PrintWriter out,
                                 TagTable tagTable,
                                 int contentModel,
                                 boolean include) {
        Enumeration tags = tagTable.getTags();
        while (tags.hasMoreElements()) {
            Dict tag = (Dict)tags.nextElement();
            boolean haveCM = ((tag.model & contentModel) != 0);
            if ((haveCM && include) || (!haveCM && !include)) {
                printTag(out, tag);
            }
        }
    }
    
    /**
     * Print a table of entries sorted by the specified content model.
     */
    private void printTagTable(PrintWriter out,
                               int contentModel) {
        TagTable tagTable = TagTable.getDefaultTagTable();

        printHeader(out, contentModel);
        printTagEntries(out, tagTable, contentModel, true);
        printTagEntries(out, tagTable, contentModel, false);
        printTrailer(out);
    }

    /**
     * TidyTagInfo program
     */
    public static void main(String[] args) throws IOException {
        if (args.length != 2) {
            System.err.println("wrong # args: TidyTagInfo sort-content-model out");
            System.exit(1);
        }
        String contentModelStr = args[0];
        String outFile = args[1];

        Integer contentModel = contentModelTable.getNumber(contentModelStr);
        if (contentModel == null) {
            System.err.println("Invalid content model: \""
                               + contentModelStr + "\"");
            Enumeration syms = contentModelTable.getSymbols();
            System.err.print("Expected one of:");
            while (syms.hasMoreElements()) {
                System.err.print (" " + (String)syms.nextElement());
            }
            System.err.println();
        }

        PrintWriter out = new PrintWriter(new FileOutputStream(outFile));
        TidyTagInfo tagInfo = new TidyTagInfo();
        tagInfo.printTagTable(out, contentModel.intValue());
        out.close();
    }
}


/**
 * Table to map bettewn and numeric id and a symbolic name.
 */
class NumberSymbolXRef {
    /**
     * Entry in the table.
     */
    private class Entry {
        public final Integer num;
        public final String sym;
        public Entry next;

        /**
         * Construct, add new entry in table.
         */
        public Entry(Integer numValue,
                     String symValue) {
            num = numValue;
            sym = symValue;
            next = null;
        }
    }

    /**
     * Number to symbol hash.
     */
    private Hashtable numToSymTable = new Hashtable();

    /**
     * Symbol to number hash.
     */
    private Hashtable symToNumTable = new Hashtable();

    /**
     * Add an entry.
     */
    public void add(int numValue,
                    String symValue) {
        Entry entry = new Entry(new Integer(numValue), symValue);
        numToSymTable.put(entry.num, entry.sym);
        symToNumTable.put(entry.sym, entry.num);
    }

    /**
     * Get an symbol given a number.
     */
    public String getSymbol(int numValue) {
        return (String)numToSymTable.get(new Integer(numValue));
    }
    
    /**
     * Get an number given a symbol.
     */
    public Integer getNumber(String symValue) {
        return (Integer)symToNumTable.get(symValue);
    }
    
    /**
     * Get an symbol given a number, error if doesn't exist.
     */
    public String retrieveSymbol(int numValue) throws XMLCError {
        String sym = (String)numToSymTable.get(new Integer(numValue));
        if (sym == null) {
            throw new XMLCError("Number not found: " + numValue);
        }
        return sym;
    }
    
    /**
     * Get an number given a symbol, error if doesn't exist.
     */
    public int retrieveNumber(String symValue) throws XMLCError {
        Integer num = (Integer)symToNumTable.get(symValue);
        if (num == null) {
            throw new XMLCError("Symbol not found: " + symValue);
        }
        return num.intValue();
    }

    /**
     * Get an enumeration of the symbols.
     */
    public Enumeration getSymbols() {
        return symToNumTable.keys();
    }

    /**
     * Get an enumeration of the nubmers.
     */
    public Enumeration getNumbers() {
        return numToSymTable.keys();
    }

    /**
     * Get the size.
     */
    public int getSize() {
        return numToSymTable.size();
    }
}

/**
 * Table of content models.
 */    
class ContentModelTable extends NumberSymbolXRef {
    /**
     * Constructor.
     */
    public ContentModelTable() {
        add(Dict.CM_EMPTY, "EMPTY");
        add(Dict.CM_HTML, "HTML");
        add(Dict.CM_HEAD, "HEAD");
        add(Dict.CM_BLOCK, "BLOCK");
        add(Dict.CM_INLINE, "INLINE");
        add(Dict.CM_LIST, "LIST");
        add(Dict.CM_DEFLIST, "DEFLIST");
        add(Dict.CM_TABLE, "TABLE");
        add(Dict.CM_ROWGRP, "ROWGRP");
        add(Dict.CM_ROW, "ROW");
        add(Dict.CM_FIELD, "FIELD");
        add(Dict.CM_OBJECT, "OBJECT");
        add(Dict.CM_PARAM, "PARAM");
        add(Dict.CM_FRAMES, "FRAMES");
        add(Dict.CM_HEADING, "HEADING");
        add(Dict.CM_OPT, "OPT");
        add(Dict.CM_IMG, "IMG");
        add(Dict.CM_MIXED, "MIXED");
        add(Dict.CM_NO_INDENT, "NO_INDENT");
        add(Dict.CM_OBSOLETE, "OBSOLETE");
        add(Dict.CM_NEW, "NEW");
    }
}
