/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: XPTest.java,v 1.1.1.1 2003/03/10 16:36:25 taweili Exp $
 */


import java.io.*;
import java.util.Enumeration;
import com.jclark.xml.parse.*;
import com.jclark.xml.parse.base.*;

public class XPTest implements Application {
    /**
     * Current indent level.
     */
    private int level = 0;

    PrintWriter out = new PrintWriter(System.out, true);

    /**
     * Indented prefix.
     */
    private void indent() {
        for (int i = 0; i < level; i++) {
            out.print("    ");
        }
    }
    
    /**
     * Indented prefix with string.
     */
    private void indent(String text) {
        indent();
        out.print(text);
    }

    /**
     * Indented prefix with string and newln.
     */
    private void indentln(String text) {
        indent();
        out.print(text);
        out.println();
    }

    /**
     * Reports the start of the document.
     * This is called once per well-formed document before any other methods.
     */
    public void startDocument() throws Exception {
        indentln("startDocument");
        level++;
    }

    /**
     * Reports the end of the prolog.
     * Called before the start of the first element.
     */
    public void endProlog(EndPrologEvent event) throws Exception {
        indentln("endProlog:");
        level++;
        printDTD(event.getDTD());
        level--;
    }

    /**
     * Reports the start of an element.
     * This includes both start-tags and empty elements.
     */
    public void startElement(StartElementEvent event) throws Exception {
        indentln("startElement:" + event.getName());
        level++;
    }

    /**
     * Reports character data.
     */
    public void characterData(CharacterDataEvent event) throws Exception {
        indentln("characterData: ref=" + event.isReference());
    }

    /**
     * Reports the end of a element.
     * This includes both end-tags and empty elements.
     */
    public void endElement(EndElementEvent event) throws Exception {
        level--;
        indentln("endElement:" + event.getName());
    }

    /**
     * Reports a processing instruction.
     * Note that processing instructions can occur before or after the
     * document element.
     */
    public void processingInstruction(ProcessingInstructionEvent event) throws Exception {
        indentln("processingInstruction: " + event.getName());
    }

    /**
     * Reports the end of the document.
     * Called once per well-formed document, after all other methods.
     * Not called if the document is not well-formed.
     */
    public void endDocument() throws Exception {
        level--;
        indentln("endDocument");
    }

    /**
     * Reports a comment.
     * Note that comments can occur before or after the
     * document element.
     */
    public void comment(CommentEvent event) throws Exception {
        indentln("comment");
    }

    /**
     * Reports the start of a CDATA section.
     */
    public void startCdataSection(StartCdataSectionEvent event) throws Exception {
        indentln("startCdataSection");
        level++;
    }

    /**
     * Reports the end of a CDATA section.
     */
    public void endCdataSection(EndCdataSectionEvent event) throws Exception {
        level--;
        indentln("endCdataSection");
    }

    /**
     * Reports the start of an entity reference.
     * This event will be followed by the result of parsing
     * the entity's replacement text.
     * This is not called for entity references in attribute values.
     */
    public void startEntityReference(StartEntityReferenceEvent event) throws Exception {
        indentln("startEntityReference: " + event.getName());
        level++;
    }

    /**
     * Reports the start of an entity reference.
     * This event follow's the result of parsing
     * the entity's replacement text.
     * This is not called for entity references in attribute values.
     */
    public void endEntityReference(EndEntityReferenceEvent event) throws Exception {
        level--;
        indentln("endEntityReference");
    }

    /**
     * Reports the start of the document type declaration.
     */
    public void startDocumentTypeDeclaration(StartDocumentTypeDeclarationEvent event) throws Exception {
        indentln("startDocumentTypeDeclaration");
        level++;
    }

    /**
     * Reports the end of the document type declaration.
     */
    public void endDocumentTypeDeclaration(EndDocumentTypeDeclarationEvent event) throws Exception {
        level--;
        indentln("endDocumentTypeDeclaration");
    }

    /**
     * Reports a markup declaration.
     */
    public void markupDeclaration(MarkupDeclarationEvent event) throws Exception {
        String which;
        switch (event.getType()) {
        case MarkupDeclarationEvent.ATTRIBUTE:
            which = "ATTRIBUTE";
            break;
        case MarkupDeclarationEvent.ELEMENT:
            which = "ELEMENT";
            break;
        case MarkupDeclarationEvent.GENERAL_ENTITY:
            which = "GENERAL_ENTITY";
            break;
        case MarkupDeclarationEvent.PARAMETER_ENTITY:
            which = "PARAMETER_ENTITY";
            break;
        case MarkupDeclarationEvent.NOTATION:
            which = "NOTATION";
            break;
        default:
            which = "*unknown*";
            break;
        }
        indentln("markupDeclaration: " + which + ": " + event.getName()
                 + ": " + event.getAttributeName());
    }

    /**
     * Print out an entity.
     */
    private void printEntity(String name,
                             Entity entity) {
        indentln("Entity: " + name
                 + " systemid=" + entity.getSystemId()
                 + " publicid=" + entity.getPublicId()
                 + " notation=" + entity.getNotationName()
                 + " base=" + entity.getBase());
    }

    /**
     * Print out a set of entities
     */
    private void printEntities(DTD dtd,
                               String title,
                               byte entityType,
                               Enumeration entities) {
        indentln("Entities: " + title);
        level++;
        while (entities.hasMoreElements()) {
            String name = (String)entities.nextElement();
            printEntity(name, dtd.getEntity(entityType, name));
        }
        level--;
    }

    /**
     * Print out the DTD.
     */
    private void printDTD(DTD dtd) {
        indentln("DTD: name=" + dtd.getDocumentTypeName()
                 + " complete=" + dtd.isComplete()
                 + " standalone= " + dtd.isStandalone());
        level++;
        printEntities(dtd, "GENERAL_ENTITY",
                      DTD.GENERAL_ENTITY,
                      dtd.entityNames(DTD.GENERAL_ENTITY));
        printEntities(dtd, "PARAMETER_ENTITY",
                      DTD.PARAMETER_ENTITY,
                      dtd.entityNames(DTD.PARAMETER_ENTITY));
        printEntities(dtd, "NOTATION",
                      DTD.NOTATION,
                      dtd.entityNames(DTD.NOTATION));
        Entity externalSubset = dtd.getEntity(DTD.PARAMETER_ENTITY, DTD.EXTERNAL_SUBSET_NAME);
        if (externalSubset == null) {
            indentln("External subset: null");
        } else {
            indentln("External subset:");
            level++;
            printEntity(DTD.EXTERNAL_SUBSET_NAME,
                        externalSubset);
            level--;
        }
        level--;
    }

    /**
     * Print a separator line.
     */
    private void printSeparator(char c) {
        for (int i = 0; i < 60; i++) {
            out.print(c);
        }
        out.println();
    }


    /**
     * Parse a document.
     */
    private void parseDoc(String doc,
                          boolean printSrc)
        throws Exception {

        printSeparator('=');
        out.println("Document = " + doc);
        printSeparator('-');
        if (printSrc) {
            String line;
            BufferedReader in = new BufferedReader(new FileReader(doc));
            while ((line = in.readLine()) != null) {
                out.println(line);
            }
            printSeparator('-');
            in.close();
        }

        Parser parser = new ParserImpl();
        parser.setApplication(this);
        parser.parseDocument(EntityManagerImpl.openFile(doc));
    }

    /**
     * Print usage message and exit.
     */
    private static void usage(String err) {
        if (err != null) {
            System.err.println(err);
        }
        System.err.println("Usage: XPTest [-printsrc] xmlfile ...");
        System.exit(1);
    }

    /**
     * Parse documents specified on command line.
     * 
     *  XPTest [-printsrc] xmlfile ...
     */
    public static void main(String args[]) throws IOException {
        boolean printSrc = false;
        int argIdx = 0;
        
        while ((argIdx < args.length) && (args[argIdx].startsWith("-"))) {
            if (args[argIdx].equals("-printsrc")) {
                printSrc = true;
            } else {
                usage("unknown option: \"" + args[argIdx] + "\"");
            }
            argIdx++;
        }
        if (argIdx == args.length) {
            usage("wrong number of arguments");
        }

        for (int idx = argIdx; idx < args.length; idx++) {
            try {
                new XPTest().parseDoc(args[idx],
                                      printSrc);
            } catch (Exception except) {
                except.printStackTrace();
            }
        }
    }
}

