/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: TestDiff.java,v 1.1.1.1 2003/03/10 16:36:25 taweili Exp $
 */

package org.enhydra.xml.driver;

import java.io.PrintWriter;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.File;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.ArrayList;
import org.enhydra.xml.driver.diff.diff;


/**
 * Object to perform comparisons of test files with expected files.
 * Also functions to search files.  Failures maybe queued to allow
 * as many files as possible to be diffed.
 */
public class TestDiff {
    /**
     * Interface for filtering lines of input.
     */
    public interface LineFilter {
        /** Filter a line */
        public String filter(String line);
    }

    /** 
     * Flag indicating that failures should not be thrown immediatly, but
     * queued for reporting later.  Allows batching of updates.
     */
    public static final int QUEUE_FAILURES = 0x01;

    /** Should missing or miss-matched expected files be updated? */
    public static final int UPDATE_EXPECTED = 0x02;
    
    /** Don't output diffs or throw exceptions, just return status  */
    public static final int NO_OUTPUT = 0x04;
    
    /** Size of buffer to use for I/O */
    private static final int BUF_SIZE = 2048;

    /** Flag set */
    private int fFlags;

    /** List of filters to apply */
    private ArrayList fFilters = new ArrayList();

    /** List of files that didn't match, or null if all have been ok */
    private ArrayList fFailedFiles;

    /** Writer for messages. */
    private PrintWriter fMsgWriter;

    /**
     * Class to filter input.  The relies on only readLine() being used.
     */
    private class LineFilterReader extends BufferedReader {
        /** File object */
        private File fFile;

        /** Constructor */
        public LineFilterReader(File inFile) throws IOException {
            super(new FileReader(inFile), BUF_SIZE);
            fFile = inFile;
        }
        
        /** Get File */
        public File getFile() {
            return fFile;
        }

        /** read a line, filtering it */
        public String readLine() throws IOException {
            String line = super.readLine();
            if (line == null) {
                return null;
            }
            int numFilters= fFilters.size();
            for (int idx = 0; idx < numFilters; idx++) {
                line = ((LineFilter)fFilters.get(idx)).filter(line);
            }
            return line;
        }
    }

    /** 
     * Constructor.
     */
    public TestDiff(PrintWriter msgWriter,
                    int flags) {
        fMsgWriter = msgWriter;
        fFlags = flags;
    }

    /** Add a filter */
    public void addFilter(LineFilter filter) {
        fFilters.add(filter);
    }

    /** Copy file lines */
    private void copyLines(BufferedReader in,
                           BufferedWriter out) throws IOException {
        String line;
        while ((line = in.readLine()) != null) {
            out.write(line, 0, line.length());
            out.newLine();
        }
    }

    /** Update an expected file, filter file contents */
    private void updateExpected(File expectedFile,
                                File gotFile) {
        fMsgWriter.println("Update " + expectedFile + " from " + gotFile);
        try {
            TestFileOps.ensureFileDir(expectedFile);
            BufferedWriter out = new BufferedWriter(new FileWriter(expectedFile),
                                                    BUF_SIZE);
            try {
                BufferedReader in = new LineFilterReader(gotFile);
                try {
                    copyLines(in, out);
                } finally {
                    in.close();
                }
            } finally {
                out.close();
            }
        } catch (IOException except) {
            throw new TestError("update of expected file failed", except);
        }
    }

    /**
     * Log a diff error.
     * @param except maybe null
     */
    private void diffError(File expectedFile,
                           File gotFile,
                           Throwable except) {
        if (fFailedFiles == null) {
            fFailedFiles = new ArrayList();
        }
        fFailedFiles.add(gotFile);

        String msg = "diff of \"" + expectedFile + "\" and \""
            + gotFile + "\" failed";
        if (except instanceof FileNotFoundException) {
            // No strack trace if one of the files is missing
            fMsgWriter.println(msg + ": " + except);
        } else if (except != null) {
            fMsgWriter.println(msg);
            except.printStackTrace();
        }
        if ((fFlags & UPDATE_EXPECTED) != 0) {
            updateExpected(expectedFile, gotFile);
        }
        if ((fFlags & QUEUE_FAILURES) == 0) {
            throw new DiffException(msg, except);
        }
    }

    /** Do the work of the diff and reporting */
    private boolean doDiff(File expectedFile,
                           BufferedReader expected,
                           LineFilterReader got) throws IOException {
        diff differ = new diff();
        differ.diffBuffer(expected, got);
        int numHunks = differ.numberOfHunk();
        if (numHunks > 0) {
            if ((fFlags & NO_OUTPUT) == 0) {
                // output requrested
                for (int idx = 0; idx < numHunks; idx++) {
                    fMsgWriter.print(differ.hunkAt(idx).convert());
                }
                fMsgWriter.println();
                diffError(expectedFile, got.getFile(), null);
            }
            return false;
        }
        return true;
    }

    /**
     * Diff an results file against an expected file, log an error
     * if the files don't match.
     * @return true if they match, false if they don't
     */
    public boolean diff(File expectedFile,
                        File gotFile) {
        try {
            BufferedReader expected
                = new BufferedReader(new FileReader(expectedFile), BUF_SIZE);
            try {
                LineFilterReader got = new LineFilterReader(gotFile);                

                try {
                    return doDiff(expectedFile, expected, got);
                } finally {
                    got.close();
                }
            } finally {
                expected.close();
            }
        } catch (IOException except) {
            diffError(expectedFile, gotFile, except);
            return false;
        }
    }

    /**
     * Generate and throw and exception if diff errors occured.  Do
     * nothing otherwise.  This allows expected files to be updated
     * for several diffs before reporting an error.
     */
    public void throwPendingDiffFailures() {
        if (fFailedFiles != null) {
            StringBuffer msg = new StringBuffer(512);
            msg.append("diff with expected files failed, see stderr for details: ");
            for (int idx = 0; idx < fFailedFiles.size(); idx++) {
                if (idx > 0) {
                    msg.append(", ");
                }
                msg.append(fFailedFiles.get(idx));
            }
            fFailedFiles = null;
            throw new DiffException(msg.toString());
        }
    }

    /**
     * Search a file for lines contining a string (not a regular expression).
     */
    public boolean fgrep(String searchStr,
                         File fileName) {
        try {
            BufferedReader in = new BufferedReader(new FileReader(fileName));
            try {
                String line;
                while ((line = in.readLine()) != null) {
                    if (line.indexOf(searchStr) >= 0) {
                        break; // found
                    }
                }
                if (line == null) {
                    return false;
                } else {
                    return true;
                }
            } finally {
                in.close();
            }
        } catch (FileNotFoundException except) {
            throw new DiffException("fgrep failed: " + except);
        } catch (IOException except) {
            throw new DiffException("fgrep of \"" + fileName + "\" failed ",
                                    except);
        }
    }

    /**
     * Generate a test error if a string is not found.
     */
    public void fgrepMustFind(String searchStr,
                              File fileName) {
        if (!fgrep(searchStr, fileName)) {
            throw new DiffException("did not find \"" + searchStr + "\" in  \""
                                    + fileName + "\"");
        }
    }

    /**
     * Generate a test error if a string is found.
     */
    public void fgrepMustNotFind(String searchStr,
                                 File fileName) {
        if (fgrep(searchStr, fileName)) {
            throw new DiffException("should not have found \"" + searchStr + "\" in  \""
                                    + fileName + "\"");
        }
    }
}
