/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DOMConverter.java,v 1.1.1.1 2003/03/10 16:37:11 taweili Exp $
 */
package org.enhydra.xml.xmlc.perf;

import java.io.*;
import org.xml.sax.SAXException;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.html.*;


/**
 * Will convert a document into a format suitable for testing.
 * All text nodes are converted into elements marked up with the
 * 'span' tag.  Each 'span' tag has an id of the format {Test[0-9]+}.
 * These span elements can be directly accessed via test classes.
 * The string representation of the converted document if printed out.
 */
public class DOMConverter {

    private int testIndex;
    private PrintStream out = System.out;
    private PrintStream err = System.err;

    /**
     * Converts the file.
     *
     * @param docClassName the class name of the DOM to be converted.
     */
    public void convert(String docClassName) {
        try {
	    Class docClass = Class.forName(docClassName);
	    XMLCFactory factory = new XMLCStdFactory(null, null);
	    Document doc = factory.create(docClass);
	    testIndex = 1;
            printNode(doc, "");
        } catch (Exception e) {
            err.println("Error: " + e.getMessage());
        }
    }

    /**
     * This will print a DOM <code>Node</code> and then recurse
     *   on its children.
     *
     * @param node <code>Node</code> object to print.
     * @param indent <code>String</code> spacing to insert 
     *               before this <code>Node</code>
     */
    public void printNode(Node node, String indent)  {

        switch (node.getNodeType()) {
	case Node.DOCUMENT_NODE:
	    NodeList nodes = node.getChildNodes();
	    if (nodes != null) {
		for (int i=0; i<nodes.getLength(); i++) {
		    printNode(nodes.item(i), "");
		}
	    }
	    break;
            
	case Node.ELEMENT_NODE:
	    String name = node.getNodeName();
	    out.print(indent + "<" + name);
	    NamedNodeMap attributes = node.getAttributes();
	    for (int i=0; i<attributes.getLength(); i++) {
		Node current = attributes.item(i);
		out.print(" " + current.getNodeName() +
				 "=\"" + current.getNodeValue() +
				 "\"");
	    }
	    out.println(">");
                
	    NodeList children = node.getChildNodes();
	    if (children != null) {
		for (int i=0; i<children.getLength(); i++) {
		    printNode(children.item(i), indent + "  ");
		}
	    }
                
	    out.println(indent + "</" + name + ">");
	    break;
            
	case Node.TEXT_NODE:
	    out.print("<SPAN ID=\"Test" + testIndex + "\">");
	    out.print(node.getNodeValue());
	    out.print("</SPAN>");
	    testIndex++;
	    break;

	case Node.CDATA_SECTION_NODE:
	    out.print(node.getNodeValue());
	    break;
            
	case Node.PROCESSING_INSTRUCTION_NODE:
	    out.println("<?" + node.getNodeName() +
			       " " + node.getNodeValue() +
			       "?>");                
	    break;
            
	case Node.ENTITY_REFERENCE_NODE:
	    out.println("&" + node.getNodeName() + ";");    
	    break;
                
	case Node.DOCUMENT_TYPE_NODE: 
	    DocumentType docType = (DocumentType)node;
	    out.print("<!DOCTYPE " + docType.getName());
	    if (docType.getPublicId() != null)  {
		out.print(" PUBLIC \"" + 
				 docType.getPublicId() + "\" ");                    
	    } else {
		out.print(" SYSTEM ");
	    }
	    out.println("\"" + docType.getSystemId() + "\">");                                
	    break;                
        }
    }
    
    /**
     * Usage: java DOMConverter domClassName
     *
     * @param args[0] the dom class name.
     */
    public static void main(String[] args) {
        if (args.length != 1) {
            System.err.println("Usage: java DOMConverter domClassName");
            System.exit(0);
        }
        
        DOMConverter converter = new DOMConverter();
        converter.convert(args[0]);
    }
    
}


