/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: SparseTextNodeEditor.java,v 1.1.1.1 2003/03/10 16:37:12 taweili Exp $
 */

package org.enhydra.xml.xmlc.perf;

import org.w3c.dom.*;
import java.lang.reflect.*;
import java.util.*;

/**
 * Edits text nodes.  May be set up to edit every every single node,
 * every other node, every fifth node, etc...  
 *
 * @see sparseOffsetList
 */
public class SparseTextNodeEditor implements DOMEditor {

    /**
     * If true, text nodes will be searched and then edited.
     * This is know to be slow with the lazydom.
     * <p>
     * If false, accessor methods of type {setTextTest[0-9]+} will
     * be called to directly edit text nodes.  This requires
     * that the DOM have been run through the DOM converter.
     *
     * @see DOMConverter
     */
    private final boolean SEARCH_EDIT = false;

    /**
     * Configure this to your test needs.  A test is run for each
     * entry in this list.  The value of an entry indicates how 'sparse'
     * the editing function is to be (i.e. 1 means edit every single
     * node, 2 means every other, 10 means every tenth, etc...)
     */
    private final int [] sparseOffsetList = new int[] {1, 2, 3, 4, 5, 10};
    private int sparseOffsetIndex = 0;
    private int count;
    private int editCount;

    /**
     * @return the current sparse offset value.
     */
    private int getSparseOffset() {
	return sparseOffsetList[sparseOffsetIndex];
    }

    /**
     * @return true if the current node should be skipped
     * (i.e. not edited)
     */
    private boolean skip(int count) {
	int offset = getSparseOffset();
	if ((count % offset) == 0) {
	    return false;
	}
	return true;
    }

    /**
     * Edits the document.
     *
     * @return the number of nodes edited.
     */
    public int edit(Document doc) throws Exception {
	count = 1;
	editCount = 0;
	if (!SEARCH_EDIT) {
	    edit(doc, getTextEditMethods(doc));
	} else {
	    edit(doc.getDocumentElement());
	}
	return editCount;
    }

    /**
     * Returns all the methods for editing text nodes.
     * The list is sorted.
     */
    private Object[] getTextEditMethods(Document doc) throws Exception {
	Method [] allMethods = doc.getClass().getMethods();
	TreeMap map = new TreeMap();
	for (int i=0; i<allMethods.length; i++) {
	    if (allMethods[i].getName().indexOf("setTextTest") > -1) {
		map.put(allMethods[i].getName(), allMethods[i]);
		
	    }
	}
	return map.values().toArray();
    }
    
    /**
     * Edits text nodes by calling direct access methods to
     * those nodes.
     * 
     * Some TEXT nodes are skipped based on the current sparse edit
     * setting.
     *
     * @see sparseOffsetList
     */
    public void edit(Document doc, Object [] methods) throws Exception {
	String [] args = new String[] {"hello"};
	for (count=0; count<methods.length; count++) {
	    if (!skip(count)) {
		((Method)methods[count]).invoke(doc, args);
		editCount++;
	    }
	}
    }

    /**
     * Recusively edits all decendant TEXT nodes of the current node.
     * Some TEXT nodes are skipped based on the current sparse edit
     * setting.
     *
     * @see sparseOffsetList
     */
    public void edit(Node node) throws Exception {
	if (node.getNodeType() == Node.TEXT_NODE) {
	    if (!skip(count)) {
		node.setNodeValue("hello");
		editCount++;
	    }
	    count++;
	} else {
	    NodeList children = node.getChildNodes();
	    for (int i=0; i<children.getLength(); i++) {
		edit(children.item(i));
	    }
	}
    }

    /**
     * String rep of editor.
     */
    public String toString() {
	int offset = getSparseOffset();
	String offsetDesc = null;
	switch (offset) {
	case 0:
	    offsetDesc = "0th";
	    break;
	case 1:
	    offsetDesc = "single";
	    break;
	case 2:
	    offsetDesc = "2nd";
	    break;
	case 3:
	    offsetDesc = "3rd";
	    break;
	default:
	    offsetDesc = offset + "th";
	    break;
	}
	return "SparseTextNodeEditor: edits every "
	    + offsetDesc + " text node in the DOM tree.";
    }

    /**
     * The edit function can be called as long as this method returns true.
     * The editor may contain multiple configurations, this method cycles
     * through each configuration.
     *
     * @return true if there is another editor configuration that should be run.
     */
    public boolean next() {
	sparseOffsetIndex++;
	if (sparseOffsetIndex >= sparseOffsetList.length) {
	    sparseOffsetIndex = sparseOffsetList.length;
	    return false;
	}
	return true;
    }

    /**
     * Reset the editor configuration queue.
     */
    public void reset() {
	sparseOffsetIndex = 0;
    }

}







